#!/usr/bin/env python3

import re
from collections import defaultdict
from pprint import pprint

import matplotlib
import mngs
import pandas as pd
import numpy as np
import warnings


class LearningCurveLogger(object):
    def __init__(
        self,
    ):
        self.logged_dict = defaultdict(dict)

        warnings.warn(
            '\n"gt_label" will be removed in the feature. Please use "true_class" instead.\n',
            DeprecationWarning,
        )

    def __call__(self, dict_to_log, step):
        """
        dict_to_log | str
            Example:
                dict_to_log = {
                    "loss_plot": float(loss),
                    "balanced_ACC_plot": float(bACC),
                    "pred_proba": pred_proba.detach().cpu().numpy(),
                    "true_class": T.cpu().numpy(),
                    "i_fold": i_fold,
                    "i_epoch": i_epoch,
                    "i_global": i_global,
                }

        step | str
            "Training", "Validation", or "Test"
        """
        ########################################
        ## delete here in the future
        ## rename "gt_label" to "true_class"
        if "gt_label" in dict_to_log.keys():
            dict_to_log["true_class"] = dict_to_log.pop("gt_label")
            # del dict_to_log["gt_label"]
        ########################################

        # assert step in ["Training", "Validation", "Test"]

        ## Initialize self.logged_dict
        for k_to_log in dict_to_log.keys():
            try:
                self.logged_dict[step][k_to_log].append(dict_to_log[k_to_log])
            except:
                self.logged_dict[step].update({k_to_log: []})
                self.logged_dict[step][k_to_log].append(dict_to_log[k_to_log])

    @property
    def dfs(self):
        return self._to_dfs_pivot(
            self.logged_dict,
            pivot_column=None,
        )

    def get_x_of_i_epoch(self, x, step, i_epoch):
        """
        lc_logger.get_x_of_i_epoch("true_label_diag", "Validation", 3)
        """
        # assert step in ['Training', "Validation", "Test"]
        indi = np.array(self.logged_dict[step]["i_epoch"]) == i_epoch
        x_all_arr = np.array(self.logged_dict[step][x])
        assert len(indi) == len(x_all_arr)
        return x_all_arr[indi]

    def plot_learning_curves(
        self,
        plt,
        plt_config_dict=None,
        title=None,
        max_n_ticks=4,
        linewidth=1,
        scattersize=50,
    ):
        """
        Plots learning curves from self.logged_dict
        """
        if plt_config_dict is not None:
            mngs.plt.configure_mpl(plt, **plt_config_dict)

        self.dfs_pivot_i_global = self._to_dfs_pivot(
            self.logged_dict, pivot_column="i_global"
        )

        ########################################
        ## Parameters
        ########################################
        COLOR_DICT = {
            "Training": "blue",
            "Validation": "green",
            "Test": "red",
        }

        keys_to_plot = self._find_keys_to_plot(self.logged_dict)

        ########################################
        ## Plots
        ########################################
        fig, axes = plt.subplots(len(keys_to_plot), 1, sharex=True, sharey=False)
        axes[-1].set_xlabel("Iteration#")
        fig.text(0.5, 0.95, title, ha="center")

        for i_plt, plt_k in enumerate(keys_to_plot):
            ax = axes[i_plt]
            ax.set_ylabel(self._rename_if_key_to_plot(plt_k))
            ax.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_ticks))
            ax.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_ticks))

            if re.search("[aA][cC][cC]", plt_k):  # acc, ylim, yticks
                ax.set_ylim(0, 1)
                ax.set_yticks([0, 0.5, 1.0])

            for step_k in self.dfs_pivot_i_global.keys():

                if step_k == re.search("^[Tt]rain", step_k):  # line
                    # if step_k == "Training":  # line
                    ax.plot(
                        self.dfs_pivot_i_global[step_k].index,
                        self.dfs_pivot_i_global[step_k][plt_k],
                        label=step_k,
                        color=mngs.plt.colors.to_RGBA(COLOR_DICT[step_k], alpha=0.9),
                        linewidth=linewidth,
                    )
                    ax.legend()

                    ########################################
                    ## Epoch starts points; just in "Training" not to b duplicated
                    ########################################
                    epoch_starts = abs(
                        self.dfs_pivot_i_global[step_k]["i_epoch"]
                        - self.dfs_pivot_i_global[step_k]["i_epoch"].shift(-1)
                    )
                    indi_global_epoch_starts = [0] + list(
                        epoch_starts[epoch_starts == 1].index
                    )

                    for i_epoch, i_global_epoch_start in enumerate(
                        indi_global_epoch_starts
                    ):
                        ax.axvline(
                            x=i_global_epoch_start,
                            ymin=-1e4,  # ax.get_ylim()[0],
                            ymax=1e4,  # ax.get_ylim()[1],
                            linestyle="--",
                            color=mngs.plt.colors.to_RGBA("gray", alpha=0.5),
                        )
                        # ax.text(
                        #     i_global_epoch_start,
                        #     ax.get_ylim()[1] * 1.0,
                        #     f"epoch#{i_epoch}",
                        # )
                    ########################################

                if (step_k == "Validation") or (step_k == "Test"):  # scatter
                    ax.scatter(
                        self.dfs_pivot_i_global[step_k].index,
                        self.dfs_pivot_i_global[step_k][plt_k],
                        label=step_k,
                        color=mngs.plt.colors.to_RGBA(COLOR_DICT[step_k], alpha=0.9),
                        s=scattersize,
                        alpha=0.9,
                    )
                    ax.legend()

        return fig

    def print(self, step):
        df_pivot_i_epoch = self._to_dfs_pivot(self.logged_dict, pivot_column="i_epoch")
        df_pivot_i_epoch_step = df_pivot_i_epoch[step]
        df_pivot_i_epoch_step.columns = self._rename_if_key_to_plot(
            df_pivot_i_epoch_step.columns
        )
        print("\n----------------------------------------\n")
        print(f"\n{step}: (mean of batches)\n")
        pprint(df_pivot_i_epoch_step)
        print("\n----------------------------------------\n")

    @staticmethod
    def _find_keys_to_plot(logged_dict):
        _steps_str = list(logged_dict.keys())
        _, keys_to_plot = mngs.general.search(
            "_plot",
            list(logged_dict[_steps_str[0]].keys()),
        )
        return keys_to_plot

    @staticmethod
    def _rename_if_key_to_plot(keys):
        def _rename_key_to_plot(key_to_plot):
            # renamed = key_to_plot[:-5]
            renamed = re.sub("_plot", "", key_to_plot).replace("_", " ")
            # renamed = renamed.replace("_", " ")
            capitalized = []
            for s in renamed.split(" "):
                if not re.search("^[A-Z]", s):
                    capitalized.append(s.capitalize())
                else:
                    capitalized.append(s)
            renamed = mngs.general.connect_strs(capitalized, filler=" ")
            return renamed

        if isinstance(keys, str):
            keys = [keys]

        out = []
        for key in keys:
            if key[-5:] == "_plot":
                out.append(_rename_key_to_plot(key))
            else:
                out.append(key)

        if len(out) == 1:
            out = out[0]

        return out

    # @staticmethod
    # def _to_dfs_pivot_i_global(logged_dict):
    #     dfs_pivot_i_global = {}
    #     for step in logged_dict.keys():
    #         df_step = mngs.pd.force_df(logged_dict[step])
    #         df_step_pvt_on_i_global = df_step.pivot_table(
    #             columns="i_global", aggfunc="mean"
    #         ).T
    #         dfs_pivot_i_global[step] = df_step_pvt_on_i_global

    #     return dfs_pivot_i_global

    @staticmethod
    def _to_dfs_pivot(logged_dict, pivot_column=None):
        dfs_pivot = {}
        for step in logged_dict.keys():
            df_step = mngs.pd.force_df(logged_dict[step])
            if pivot_column is not None:
                dfs_pivot[step] = df_step.pivot_table(
                    columns=pivot_column, aggfunc="mean"
                ).T
            else:
                dfs_pivot[step] = df_step

        return dfs_pivot


if __name__ == "__main__":
    import warnings

    import matplotlib.pyplot as plt
    import torch
    import torch.nn as nn
    from sklearn.metrics import balanced_accuracy_score
    from torch.utils.data import DataLoader, TensorDataset
    from torch.utils.data.dataset import Subset
    from torchvision import datasets

    import sys

    ################################################################################
    ## Sets tee
    ################################################################################
    sdir = mngs.io.path.mk_spath("")  # "/tmp/sdir/"
    sys.stdout, sys.stderr = mngs.general.tee(sys, sdir)

    ################################################################################
    ## NN
    ################################################################################
    class Perceptron(nn.Module):
        def __init__(self):
            super().__init__()
            self.l1 = nn.Linear(28 * 28, 50)
            self.l2 = nn.Linear(50, 10)

        def forward(self, x):
            x = x.view(-1, 28 * 28)
            x = self.l1(x)
            x = self.l2(x)
            return x

    ################################################################################
    ## Prepaires demo data
    ################################################################################
    ## Downloads
    _ds_tra_val = datasets.MNIST("/tmp/mnist", train=True, download=True)
    _ds_tes = datasets.MNIST("/tmp/mnist", train=False, download=True)

    ## Training-Validation splitting
    n_samples = len(_ds_tra_val)  # n_samples is 60000
    train_size = int(n_samples * 0.8)  # train_size is 48000

    subset1_indices = list(range(0, train_size))  # [0,1,.....47999]
    subset2_indices = list(range(train_size, n_samples))  # [48000,48001,.....59999]

    _ds_tra = Subset(_ds_tra_val, subset1_indices)
    _ds_val = Subset(_ds_tra_val, subset2_indices)

    ## to tensors
    ds_tra = TensorDataset(
        _ds_tra.dataset.data.to(torch.float32),
        _ds_tra.dataset.targets,
    )
    ds_val = TensorDataset(
        _ds_val.dataset.data.to(torch.float32),
        _ds_val.dataset.targets,
    )
    ds_tes = TensorDataset(
        _ds_tes.data.to(torch.float32),
        _ds_tes.targets,
    )

    ## to dataloaders
    batch_size = 64
    dl_tra = DataLoader(
        dataset=ds_tra,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
    )

    dl_val = DataLoader(
        dataset=ds_val,
        batch_size=batch_size,
        shuffle=False,
        drop_last=True,
    )

    dl_tes = DataLoader(
        dataset=ds_tes,
        batch_size=batch_size,
        shuffle=False,
        drop_last=True,
    )

    ################################################################################
    ## Preparation
    ################################################################################
    model = Perceptron()
    loss_func = nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)
    softmax = nn.Softmax(dim=-1)

    ################################################################################
    ## Main
    ################################################################################
    lc_logger = LearningCurveLogger()
    i_global = 0

    n_classes = len(dl_tra.dataset.tensors[1].unique())
    i_fold = 0
    max_epochs = 3

    for i_epoch in range(max_epochs):
        step = "Validation"
        for i_batch, batch in enumerate(dl_val):

            X, T = batch
            logits = model(X)
            pred_proba = softmax(logits)
            pred_class = pred_proba.argmax(dim=-1)
            loss = loss_func(logits, T)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                bACC = balanced_accuracy_score(T, pred_class)

            dict_to_log = {
                "loss_plot": float(loss),
                "balanced_ACC_plot": float(bACC),
                "pred_proba": pred_proba.detach().cpu().numpy(),
                "gt_label": T.cpu().numpy(),
                # "true_class": T.cpu().numpy(),
                "i_fold": i_fold,
                "i_epoch": i_epoch,
                "i_global": i_global,
            }
            lc_logger(dict_to_log, step)

        lc_logger.print(step)

        step = "Training"
        for i_batch, batch in enumerate(dl_tra):
            optimizer.zero_grad()

            X, T = batch
            logits = model(X)
            pred_proba = softmax(logits)
            pred_class = pred_proba.argmax(dim=-1)
            loss = loss_func(logits, T)

            loss.backward()
            optimizer.step()

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                bACC = balanced_accuracy_score(T, pred_class)

            dict_to_log = {
                "loss_plot": float(loss),
                "balanced_ACC_plot": float(bACC),
                "pred_proba": pred_proba.detach().cpu().numpy(),
                "gt_label": T.cpu().numpy(),
                # "true_class": T.cpu().numpy(),
                "i_fold": i_fold,
                "i_epoch": i_epoch,
                "i_global": i_global,
            }
            lc_logger(dict_to_log, step)

            i_global += 1

        lc_logger.print(step)

    step = "Test"
    for i_batch, batch in enumerate(dl_tes):

        X, T = batch
        logits = model(X)
        pred_proba = softmax(logits)
        pred_class = pred_proba.argmax(dim=-1)
        loss = loss_func(logits, T)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            bACC = balanced_accuracy_score(T, pred_class)

        dict_to_log = {
            "loss_plot": float(loss),
            "balanced_ACC_plot": float(bACC),
            "pred_proba": pred_proba.detach().cpu().numpy(),
            # "gt_label": T.cpu().numpy(),
            "true_class": T.cpu().numpy(),
            "i_fold": i_fold,
            "i_epoch": i_epoch,
            "i_global": i_global,
        }
        lc_logger(dict_to_log, step)

    lc_logger.print(step)

    plt_config_dict = dict(
        # figsize=(8.7, 10),
        figscale=2.5,
        labelsize=16,
        fontsize=12,
        legendfontsize=12,
        tick_size=0.8,
        tick_width=0.2,
    )

    fig = lc_logger.plot_learning_curves(
        plt,
        plt_config_dict=plt_config_dict,
        title=f"fold#{i_fold}",
        linewidth=1,
        scattersize=50,
    )
    fig.show()
    # mngs.general.save(fig, sdir + f"fold#{i_fold}.png")



#!/usr/bin/env python3

# import warnings

# try:
#     from ._get_proc_usages import get_proc_usages
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import get_proc_usages from ._get_proc_usages.")

# try:
#     from ._get_specs import get_specs
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import get_specs from ._get_specs.")

# try:
#     from ._rec_procs import rec_procs
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import rec_procs from ._rec_procs.")

# try:
#     from . import _utils
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import _utils.")


from ._get_proc_usages import get_proc_usages
from ._get_specs import get_specs
from ._rec_procs import rec_procs
from . import _utils

_ = None  # keep the importing orders

# from .limit_RAM import get_RAM, limit_RAM

# from .get import system_info

## EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-05 06:01:32 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/plt/ax/_fill_between_v.py


"""
This script does XYZ.
"""


"""
Imports
"""

import numpy as np


def fillv(axes, starts, ends, color="red", alpha=0.2):
    """
    Fill between specified start and end intervals on an axis or array of axes.

    Parameters
    ----------
    axes : matplotlib.axes.Axes or numpy.ndarray of matplotlib.axes.Axes
        The axis object(s) to fill intervals on.
    starts : array-like
        Array-like of start positions.
    ends : array-like
        Array-like of end positions.
    color : str, optional
        The color to use for the filled regions. Default is "red".
    alpha : float, optional
        The alpha blending value, between 0 (transparent) and 1 (opaque). Default is 0.2.

    Returns
    -------
    list
        List of axes with filled intervals.
    """

    is_axes = isinstance(axes, np.ndarray)

    axes = axes if isinstance(axes, np.ndarray) else [axes]

    for ax in axes:
        for start, end in zip(starts, ends):
            ax.axvspan(start, end, color=color, alpha=alpha)

    if not is_axes:
        return axes[0]
    else:
        return axes


# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 18:39:54 (ywatanabe)"

# try:
#     from ._AxiswiseDropout import AxiswiseDropout
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import AxiswiseDropout.")

# try:
#     from ._BNet import BNet, BNet_config
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._BNet.")

# try:
#     from ._ChannelGainChanger import ChannelGainChanger
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import ChannelGainChanger.")

# try:
#     from ._DropoutChannels import DropoutChannels
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import DropoutChannels.")

# try:
#     from ._Filters import (
#         BandPassFilter,
#         BandStopFilter,
#         DifferentiableBandPassFilter,
#         GaussianFilter,
#         HighPassFilter,
#         LowPassFilter,
#     )
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._Filters.")

# try:
#     from ._FreqGainChanger import FreqGainChanger
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import FreqGainChanger.")

# try:
#     from ._Hilbert import Hilbert
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import Hilbert.")

# try:
#     from ._MNet_1000 import MNet_1000, MNet_config
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._MNet_1000.")

# try:
#     from ._ModulationIndex import ModulationIndex
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import ModulationIndex.")

# try:
#     from ._PAC import PAC
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import PAC.")

# try:
#     from ._PSD import PSD
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import PSD.")

# try:
#     from ._ResNet1D import ResNet1D, ResNetBasicBlock
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._ResNet1D.")

# try:
#     from ._SpatialAttention import SpatialAttention
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import SpatialAttention.")

# try:
#     from ._SwapChannels import SwapChannels
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import SwapChannels.")

# try:
#     from ._TransposeLayer import TransposeLayer
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import TransposeLayer.")

# try:
#     from ._Wavelet import Wavelet
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import Wavelet.")

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-27 16:55:44 (ywatanabe)"

from ._AxiswiseDropout import AxiswiseDropout
from ._BNet import BNet, BNet_config
from ._ChannelGainChanger import ChannelGainChanger
from ._DropoutChannels import DropoutChannels
from ._Filters import (
    BandPassFilter,
    BandStopFilter,
    DifferentiableBandPassFilter,
    GaussianFilter,
    HighPassFilter,
    LowPassFilter,
)
from ._FreqGainChanger import FreqGainChanger
from ._Hilbert import Hilbert

# from ._FreqDropout import FreqDropout
from ._MNet_1000 import MNet_1000, MNet_config
from ._ModulationIndex import ModulationIndex
from ._PAC import PAC

# from ._PAC_dev import PAC_dev
from ._PSD import PSD
from ._ResNet1D import ResNet1D, ResNetBasicBlock
from ._SpatialAttention import SpatialAttention
from ._SwapChannels import SwapChannels
from ._TransposeLayer import TransposeLayer
from ._Wavelet import Wavelet



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-13 20:39:12 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_ChatGPT.py


"""Imports"""
import os
import sys

import google.generativeai as genai
import matplotlib.pyplot as plt
import mngs
from mngs.ai._gen_ai._BaseGenAI import BaseGenAI

"""Functions & Classes"""


class Google(BaseGenAI):
    def __init__(
        self,
        system_setting="",
        api_key=os.getenv("GOOGLE_API_KEY"),
        model="gemini-1.5-pro-latest",
        stream=False,
        seed=None,
        n_keep=1,
        temperature=1.0,
        chat_history=None,
        max_tokens=4096,
    ):

        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key=api_key,
            stream=stream,
            seed=seed,
            n_keep=n_keep,
            temperature=temperature,
            provider="Gemini",
            chat_history=chat_history,
            max_tokens=max_tokens,
        )

        genai.configure(api_key=self.api_key)
        # genai.configure(api_key=self.api_key)

    def _init_client(
        self,
    ):
        genai.configure(api_key=self.api_key)
        generation_config = genai.GenerationConfig(
            temperature=self.temperature,
        )
        client = genai.GenerativeModel(
            self.model, generation_config=generation_config
        )
        chat_client = client.start_chat(history=self.history)
        return chat_client

    def _api_call_static(
        self,
    ):
        prompt = self.history[-1]["content"]
        response = self.client.send_message(prompt)
        self.input_tokens += response.usage_metadata.prompt_token_count
        self.output_tokens += response.usage_metadata.candidates_token_count
        out_text = response.text
        return out_text

    # def _api_call_static(self):
    #     prompt = self.history[-1]["content"]
    #     response = self.client.send_message(prompt)
    #     if response.usage_metadata:
    #         self.input_tokens += response.usage_metadata.prompt_token_count
    #         self.output_tokens += (
    #             response.usage_metadata.candidates_token_count
    #         )
    #     out_text = response.text
    #     return out_text

    def _api_call_stream(
        self,
    ):
        prompt = self.history[-1]["content"]
        responses = self.client.send_message(prompt, stream=True)
        for chunk in responses:
            if chunk:
                try:
                    self.input_tokens += (
                        chunk.usage_metadata.prompt_token_count
                    )
                except:
                    pass
                try:
                    self.output_tokens += (
                        chunk.usage_metadata.candidates_token_count
                    )
                except:
                    pass

                yield chunk.text

    # def _api_call_stream(self):
    #     prompt = self.history[-1]["content"]
    #     responses = self.client.send_message(prompt, stream=True)
    #     for chunk in responses:
    #         if chunk:
    #             if chunk.usage_metadata:
    #                 self.input_tokens += (
    #                     chunk.usage_metadata.prompt_token_count
    #                 )
    #                 self.output_tokens += (
    #                     chunk.usage_metadata.candidates_token_count
    #                 )
    #             yield chunk.text


def main():
    # ai = mngs.ai.GenAI("gpt-4o")
    # ai = mngs.ai.GenAI("gemini-1.5-pro-latest", stream=False)
    ai = mngs.ai.GenAI("gemini-1.5-flash-latest", stream=False)
    # ai = mngs.ai.GenAI("gpt-4o", stream=False)
    print(ai("hi"))


if __name__ == "__main__":
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3

try:
    from ._bonferroni_correction import bonferroni_correction
except ImportError as e:
    warnings.warn(f"Warning: Failed to import bonferroni_correction from ._bonferroni_correction.")

try:
    from ._brunner_munzel_test import brunner_munzel_test
except ImportError as e:
    warnings.warn(f"Warning: Failed to import brunner_munzel_test from ._brunner_munzel_test.")

try:
    from ._calc_partial_corr import calc_partial_corr
except ImportError as e:
    warnings.warn(f"Warning: Failed to import calc_partial_corr from ._calc_partial_corr.")

try:
    from ._corr_test import corr_test
except ImportError as e:
    warnings.warn(f"Warning: Failed to import corr_test from ._corr_test.")

try:
    from ._fdr_correction import fdr_correction
except ImportError as e:
    warnings.warn(f"Warning: Failed to import fdr_correction from ._fdr_correction.")

try:
    from ._general import kurtosis, mean, median, q, skewness, std, zscore
except ImportError as e:
    warnings.warn(f"Warning: Failed to import from ._general.")

try:
    from ._multicompair import multicompair
except ImportError as e:
    warnings.warn(f"Warning: Failed to import multicompair from ._multicompair.")

try:
    from ._nocorrelation_test import nocorrelation_test
except ImportError as e:
    warnings.warn(f"Warning: Failed to import nocorrelation_test from ._nocorrelation_test.")

try:
    from ._p2stars import p2stars
except ImportError as e:
    warnings.warn(f"Warning: Failed to import p2stars from ._p2stars.")

try:
    from ._smirnov_grubbs import smirnov_grubbs
except ImportError as e:
    warnings.warn(f"Warning: Failed to import smirnov_grubbs from ._smirnov_grubbs.")

try:
    from ._find_pval import find_pval
except ImportError as e:
    warnings.warn(f"Warning: Failed to import find_pval from ._find_pval.")


# from ._bonferroni_correction import bonferroni_correction
# from ._brunner_munzel_test import brunner_munzel_test
# from ._calc_partial_corr import calc_partial_corr
# from ._corr_test import corr_test
# from ._fdr_correction import fdr_correction#, fdr_correction_torch
# from ._general import kurtosis, mean, median, q, skewness, std, zscore
# from ._multicompair import multicompair
# from ._nocorrelation_test import nocorrelation_test
# from ._p2stars import p2stars
# from ._smirnov_grubbs import smirnov_grubbs
# from ._find_pval import find_pval



#!/usr/bin/env python3

import os
import random
import sys
from collections import defaultdict
from glob import glob
from pprint import pprint

import matplotlib
import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd
import torch
from sklearn.metrics import (balanced_accuracy_score, classification_report,
                             confusion_matrix, matthews_corrcoef)


class MultiClassificationReporter(object):
    def __init__(self, sdir, tgts=None):
        if tgts is None:
            sdirs = [""]
        else:
            sdirs = [os.path.join(sdir, tgt, "/") for tgt in tgts]
        sdirs = [sdir + tgt + "/" for tgt in tgts]

        self.tgt2id = {tgt: i_tgt for i_tgt, tgt in enumerate(tgts)}
        self.reporters = [ClassificationReporter(sdir) for sdir in sdirs]

    def add(self, obj_name, obj, tgt=None):
        i_tgt = self.tgt2id[tgt]
        self.reporters[i_tgt].add(obj_name, obj)

    def calc_metrics(
        self,
        true_class,
        pred_class,
        pred_proba,
        labels=None,
        i_fold=None,
        show=True,
        auc_plt_config=dict(
            figsize=(7, 7),
            labelsize=8,
            fontsize=7,
            legendfontsize=6,
            tick_size=0.8,
            tick_width=0.2,
        ),
        tgt=None,
    ):
        i_tgt = self.tgt2id[tgt]
        self.reporters[i_tgt].calc_metrics(
            true_class,
            pred_class,
            pred_proba,
            labels=labels,
            i_fold=i_fold,
            show=show,
            auc_plt_config=auc_plt_config,
        )

    def summarize(
        self,
        n_round=3,
        show=False,
        tgt=None,
    ):
        i_tgt = self.tgt2id[tgt]
        self.reporters[i_tgt].summarize(
            n_round=n_round,
            show=show,
        )

    def save(
        self,
        files_to_reproduce=None,
        meta_dict=None,
        tgt=None,
    ):
        i_tgt = self.tgt2id[tgt]
        self.reporters[i_tgt].save(
            files_to_reproduce=files_to_reproduce,
            meta_dict=meta_dict,
        )

    def plot_and_save_conf_mats(
        self,
        plt,
        extend_ratio=1.0,
        colorbar=True,
        confmat_plt_config=None,
        sci_notation_kwargs=None,
        tgt=None,
    ):
        i_tgt = self.tgt2id[tgt]
        self.reporters[i_tgt].plot_and_save_conf_mats(
            plt,
            extend_ratio=extend_ratio,
            colorbar=colorbar,
            confmat_plt_config=confmat_plt_config,
            sci_notation_kwargs=sci_notation_kwargs,
        )


class ClassificationReporter(object):
    """Saves the following metrics under sdir.
       - Balanced Accuracy
       - MCC
       - Confusion Matrix
       - Classification Report
       - ROC AUC score / curve
       - PRE-REC AUC score / curve

    Example is described in this file.
    """

    def __init__(self, sdir):
        self.sdir = sdir
        self.folds_dict = defaultdict(list)
        mngs.general.fix_seeds(
            os=os, random=random, np=np, torch=torch, show=False
        )

    def add(
        self,
        obj_name,
        obj,
    ):
        """
        ## fig
        fig, ax = plt.subplots()
        ax.plot(np.random.rand(10))
        reporter.add("manu_figs", fig)

        ## DataFrame
        df = pd.DataFrame(np.random.rand(5, 3))
        reporter.add("manu_dfs", df)

        ## scalar
        scalar = random.random()
        reporter.add("manu_scalers", scalar)
        """
        assert isinstance(obj_name, str)
        self.folds_dict[obj_name].append(obj)

    @staticmethod
    def calc_bACC(true_class, pred_class, i_fold, show=False):
        """Balanced ACC"""
        balanced_acc = balanced_accuracy_score(true_class, pred_class)
        if show:
            print(f"\nBalanced ACC in fold#{i_fold} was {balanced_acc:.3f}\n")
        return balanced_acc

    @staticmethod
    def calc_mcc(true_class, pred_class, i_fold, show=False):
        """MCC"""
        mcc = float(matthews_corrcoef(true_class, pred_class))
        if show:
            print(f"\nMCC in fold#{i_fold} was {mcc:.3f}\n")
        return mcc

    @staticmethod
    def calc_conf_mat(true_class, pred_class, labels, i_fold, show=False):
        """
        Confusion Matrix
        This method assumes unique classes of true_class and pred_class are the same.
        """
        # conf_mat = pd.DataFrame(
        #     data=confusion_matrix(true_class, pred_class),
        #     columns=pred_labels,
        #     index=true_labels,
        # )

        conf_mat = pd.DataFrame(
            data=confusion_matrix(
                true_class, pred_class, labels=np.arange(len(labels))
            ),
            columns=labels,
        ).set_index(pd.Series(list(labels)))

        if show:
            print(f"\nConfusion Matrix in fold#{i_fold}: \n")
            pprint(conf_mat)
            print()

        return conf_mat

    @staticmethod
    def calc_clf_report(
        true_class, pred_class, labels, balanced_acc, i_fold, show=False
    ):
        """Classification Report"""
        clf_report = pd.DataFrame(
            classification_report(
                true_class,
                pred_class,
                labels=np.arange(len(labels)),
                target_names=labels,
                output_dict=True,
            )
        )

        # ACC to bACC
        clf_report["accuracy"] = balanced_acc
        clf_report = pd.concat(
            [
                clf_report[labels],
                clf_report[["accuracy", "macro avg", "weighted avg"]],
            ],
            axis=1,
        )
        clf_report = clf_report.rename(
            columns={"accuracy": "balanced accuracy"}
        )
        clf_report = clf_report.round(3)
        # Renames 'support' to 'sample size'
        clf_report["index"] = clf_report.index
        clf_report.loc["support", "index"] = "sample size"
        clf_report.set_index("index", drop=True, inplace=True)
        clf_report.index.name = None
        if show:
            print(f"\nClassification Report for fold#{i_fold}:\n")
            pprint(clf_report)
            print()
        return clf_report

    @staticmethod
    def calc_and_plot_roc_curve(
        true_class, pred_proba, labels, sdir_for_csv=None
    ):
        # ROC-AUC
        fig_roc, metrics_roc_auc_dict = mngs.ml.plt.roc_auc(
            plt,
            true_class,
            pred_proba,
            labels,
            sdir_for_csv=sdir_for_csv,
        )
        plt.close()
        return fig_roc, metrics_roc_auc_dict

    @staticmethod
    def calc_and_plot_pre_rec_curve(true_class, pred_proba, labels):
        # PRE-REC AUC
        fig_pre_rec, metrics_pre_rec_auc_dict = mngs.ml.plt.pre_rec_auc(
            plt, true_class, pred_proba, labels
        )
        plt.close()
        return fig_pre_rec, metrics_pre_rec_auc_dict

    def calc_metrics(
        self,
        true_class,
        pred_class,
        pred_proba,
        labels=None,
        i_fold=None,
        show=True,
        auc_plt_config=dict(
            figsize=(7, 7),
            labelsize=8,
            fontsize=7,
            legendfontsize=6,
            tick_size=0.8,
            tick_width=0.2,
        ),
    ):
        """
        Calculates ACC, Confusion Matrix, Classification Report, and ROC-AUC score on a fold.
        Metrics and curves will be kept in self.folds_dict.
        """

        ## Preparation
        # for convenience
        true_class = (
            mngs.general.torch_to_arr(true_class).astype(int).reshape(-1)
        )
        pred_class = (
            mngs.general.torch_to_arr(pred_class)
            .astype(np.float64)
            .reshape(-1)
        )
        pred_proba = mngs.general.torch_to_arr(pred_proba).astype(np.float64)

        # for curves
        mngs.plt.configure_mpl(
            plt,
            **auc_plt_config,
        )

        ## Calc metrics
        # Balanced ACC
        bacc = self.calc_bACC(true_class, pred_class, i_fold, show=show)
        self.folds_dict["balanced_acc"].append(bacc)

        # MCC
        self.folds_dict["mcc"].append(
            self.calc_mcc(true_class, pred_class, i_fold, show=show)
        )

        # Confusion Matrix
        self.folds_dict["conf_mat/conf_mat"].append(
            self.calc_conf_mat(
                true_class,
                pred_class,
                labels,
                i_fold,
                show=show,
            )
        )

        # Classification Report
        self.folds_dict["clf_report"].append(
            self.calc_clf_report(
                true_class, pred_class, labels, bacc, i_fold, show=show
            )
        )

        ## Curves
        # ROC curve
        self.sdir_for_roc_csv = f"{self.sdir}roc/csv/"
        fig_roc, metrics_roc_auc_dict = self.calc_and_plot_roc_curve(
            true_class,
            pred_proba,
            labels,
            sdir_for_csv=self.sdir_for_roc_csv + f"fold#{i_fold}/",
        )
        self.folds_dict["roc/micro"].append(
            metrics_roc_auc_dict["roc_auc"]["micro"]
        )
        self.folds_dict["roc/macro"].append(
            metrics_roc_auc_dict["roc_auc"]["macro"]
        )
        self.folds_dict["roc/figs"].append(fig_roc)

        # PRE-REC curve
        fig_pre_rec, metrics_pre_rec_auc_dict = (
            self.calc_and_plot_pre_rec_curve(true_class, pred_proba, labels)
        )
        self.folds_dict["pre_rec/micro"].append(
            metrics_pre_rec_auc_dict["pre_rec_auc"]["micro"]
        )
        self.folds_dict["pre_rec/macro"].append(
            metrics_pre_rec_auc_dict["pre_rec_auc"]["macro"]
        )
        self.folds_dict["pre_rec/figs"].append(fig_pre_rec)

    @staticmethod
    def _mk_cv_index(n_folds):
        return [
            f"{n_folds}-folds_CV_mean",
            f"{n_folds}-fold_CV_std",
        ] + [f"fold#{i_fold}" for i_fold in range(n_folds)]

    def summarize_roc(
        self,
    ):

        folds_dirs = glob(self.sdir_for_roc_csv + "fold#*")
        n_folds = len(folds_dirs)

        # get class names
        _csv_files = glob(os.path.join(folds_dirs[0], "*"))
        classes_str = [
            csv_file.split("/")[-1].split(".csv")[0] for csv_file in _csv_files
        ]

        # dfs_classes = []
        # take mean and std by each class
        for cls_str in classes_str:

            fpaths_cls = [
                os.path.join(fold_dir, f"{cls_str}.csv")
                for fold_dir in folds_dirs
            ]

            ys = []
            roc_aucs = []
            for fpath_cls in fpaths_cls:
                loaded_df = mngs.io.load(fpath_cls)
                ys.append(loaded_df["y"])
                roc_aucs.append(loaded_df["roc_auc"])
            ys = pd.concat(ys, axis=1)
            roc_aucs = pd.concat(roc_aucs, axis=1)

            df_cls = loaded_df[["x"]].copy()
            df_cls["y_mean"] = ys.mean(axis=1)
            df_cls["y_std"] = ys.std(axis=1)
            df_cls["roc_auc_mean"] = roc_aucs.mean(axis=1)
            df_cls["roc_auc_std"] = roc_aucs.std(axis=1)

            spath_cls = os.path.join(
                self.sdir_for_roc_csv, f"k-fold_mean_std/{cls_str}.csv"
            )
            mngs.io.save(df_cls, spath_cls)
            # dfs_classes.append(df_cls)

    def summarize(
        self,
        n_round=3,
        show=False,
    ):
        """
        1) Take mean and std of scalars/pd.Dataframes for folds.
        2) Replace self.folds_dict with the summarized DataFrames.
        """
        self.summarize_roc()

        _n_folds_all = [
            len(self.folds_dict[k]) for k in self.folds_dict.keys()
        ]  # sometimes includes 0 because AUC curves are not always defined.
        self.n_folds_intended = max(_n_folds_all)

        for i_k, k in enumerate(self.folds_dict.keys()):
            n_folds = _n_folds_all[i_k]

            if n_folds != 0:
                ## listed scalars
                if mngs.general.is_listed_X(self.folds_dict[k], [float, int]):
                    mm = np.mean(self.folds_dict[k])
                    ss = np.std(self.folds_dict[k], ddof=1)
                    sr = pd.DataFrame(
                        data=[mm, ss] + self.folds_dict[k],
                        index=self._mk_cv_index(n_folds),
                        columns=[k],
                    )
                    self.folds_dict[k] = sr.round(n_round)

                ## listed pd.DataFrames
                elif mngs.general.is_listed_X(
                    self.folds_dict[k], pd.DataFrame
                ):
                    zero_df_for_mm = 0 * self.folds_dict[k][0].copy()
                    zero_df_for_ss = 0 * self.folds_dict[k][0].copy()

                    mm = (
                        zero_df_for_mm
                        + np.stack(self.folds_dict[k]).mean(axis=0)
                    ).round(n_round)

                    ss = (
                        zero_df_for_ss
                        + np.stack(self.folds_dict[k]).std(axis=0, ddof=1)
                    ).round(n_round)

                    self.folds_dict[k] = [mm, ss] + [
                        df_fold.round(n_round)
                        for df_fold in self.folds_dict[k]
                    ]

                    if show:
                        print(
                            "\n----------------------------------------\n"
                            f"\n{k}\n"
                            f"\n{n_folds}-fold-CV mean:\n"
                        )
                        pprint(self.folds_dict[k][0])
                        print(f"\n\n{n_folds}-fold-CV std.:\n")
                        pprint(self.folds_dict[k][1])
                        print("\n\n----------------------------------------\n")

                ## listed figures
                elif mngs.general.is_listed_X(
                    self.folds_dict[k], matplotlib.figure.Figure
                ):
                    pass

                else:
                    print(f"{k} was not summarized")
                    print(type(self.folds_dict[k][0]))

    def save(
        self,
        files_to_reproduce=None,
        meta_dict=None,
    ):
        """
        1) Saves the content of self.folds_dict.
        2) Plots the colormap of confusion matrices and saves them.
        3) Saves passed meta_dict under self.sdir

        Example:
            meta_df_1 = pd.DataFrame(data=np.random.rand(3,3))
            meta_dict_1 = {"a": 0}
            meta_dict_2 = {"b": 0}
            meta_dict = {"meta_1.csv": meta_df_1,
                         "meta_1.yaml": meta_dict_1,
                         "meta_2.yaml": meta_dict_1,
            }

        """
        if meta_dict is not None:
            for k, v in meta_dict.items():
                mngs.io.save(v, self.sdir + k)

        for k in self.folds_dict.keys():

            ## pd.Series / pd.DataFrame
            if isinstance(self.folds_dict[k], pd.Series) or isinstance(
                self.folds_dict[k], pd.DataFrame
            ):
                mngs.io.save(self.folds_dict[k], self.sdir + f"{k}.csv")

            ## listed pd.DataFrame
            elif mngs.general.is_listed_X(self.folds_dict[k], pd.DataFrame):
                mngs.io.save(
                    self.folds_dict[k],
                    self.sdir + f"{k}.csv",
                    # indi_suffix=self.cv_index,
                    indi_suffix=self._mk_cv_index(len(self.folds_dict[k])),
                )

            ## listed figures
            elif mngs.general.is_listed_X(
                self.folds_dict[k], matplotlib.figure.Figure
            ):
                for i_fold, fig in enumerate(self.folds_dict[k]):
                    mngs.io.save(
                        self.folds_dict[k][i_fold],
                        self.sdir + f"{k}/fold#{i_fold}.png",
                    )

            else:
                print(f"{k} was not saved")
                print(type(self.folds_dict[k]))

        if files_to_reproduce is not None:
            if isinstance(files_to_reproduce, list):
                files_to_reproduce = [files_to_reproduce]
            for f in files_to_reproduce:
                mngs.io.save(f, self.sdir)

    def plot_and_save_conf_mats(
        self,
        plt,
        extend_ratio=1.0,
        colorbar=True,
        confmat_plt_config=None,
        sci_notation_kwargs=None,
    ):
        def _inner_plot_conf_mat(
            plt,
            cm_df,
            title,
            extend_ratio=1.0,
            colorbar=True,
            sci_notation_kwargs=None,
        ):
            labels = list(cm_df.columns)
            fig_conf_mat = mngs.ml.plt.confusion_matrix(
                plt,
                cm_df.T,
                labels=labels,
                title=title,
                x_extend_ratio=extend_ratio,
                y_extend_ratio=extend_ratio,
                colorbar=colorbar,
            )

            if sci_notation_kwargs is not None:
                fig_conf_mat.axes[-1] = mngs.plt.ax_scientific_notation(
                    fig_conf_mat.axes[-1], **sci_notation_kwargs
                )
            return fig_conf_mat

        ## Configures mpl
        mngs.plt.configure_mpl(
            plt,
            **confmat_plt_config,
        )

        ########################################
        ## Prepares confmats dfs
        ########################################
        ## Drops mean and std for the folds
        try:
            conf_mats = self.folds_dict["conf_mat/conf_mat"][
                -self.n_folds_intended :
            ]

        except Exception as e:
            print(e)
            conf_mats = self.folds_dict["conf_mat/conf_mat"]

        ## Prepaires conf_mat_overall_sum
        conf_mat_zero = 0 * conf_mats[0].copy()  # get the table format
        conf_mat_overall_sum = conf_mat_zero + np.stack(conf_mats).sum(axis=0)

        ########################################
        ## Plots & Saves
        ########################################
        # each fold's conf
        for i_fold, cm in enumerate(conf_mats):
            title = f"Test fold#{i_fold}"
            fig_conf_mat_fold = _inner_plot_conf_mat(
                plt,
                cm,
                title,
                extend_ratio=extend_ratio,
                colorbar=colorbar,
                sci_notation_kwargs=sci_notation_kwargs,
            )
            mngs.io.save(
                fig_conf_mat_fold,
                self.sdir + f"conf_mat/figs/fold#{i_fold}.png",
            )
            plt.close()

        ## overall_sum conf_mat
        title = f"{self.n_folds_intended}-CV overall sum"
        fig_conf_mat_overall_sum = _inner_plot_conf_mat(
            plt,
            conf_mat_overall_sum,
            title,
            extend_ratio=extend_ratio,
            colorbar=colorbar,
            sci_notation_kwargs=sci_notation_kwargs,
        )
        mngs.io.save(
            fig_conf_mat_overall_sum,
            self.sdir
            + f"conf_mat/figs/{self.n_folds_intended}-fold_cv_overall-sum.png",
        )
        plt.close()


if __name__ == "__main__":
    import random
    import sys

    import mngs
    import numpy as np
    from catboost import CatBoostClassifier, Pool
    from sklearn.datasets import load_digits
    from sklearn.model_selection import StratifiedKFold

    ################################################################################
    ## Sets tee
    ################################################################################
    sdir = mngs.io.mk_spath(
        "./tmp/sdir-ClassificationReporter/"
    )  # "/tmp/sdir/"
    sys.stdout, sys.stderr = mngs.general.tee(sys, sdir)

    ################################################################################
    ## Fixes seeds
    ################################################################################
    mngs.general.fix_seeds(np=np)

    ## Loads
    mnist = load_digits()
    X, T = mnist.data, mnist.target
    labels = mnist.target_names.astype(str)

    ## Main
    skf = StratifiedKFold(n_splits=5, shuffle=True)
    # reporter = ClassificationReporter(sdir)
    mreporter = MultiClassificationReporter(sdir, tgts=["Test1", "Test2"])
    for i_fold, (indi_tra, indi_tes) in enumerate(skf.split(X, T)):
        X_tra, T_tra = X[indi_tra], T[indi_tra]
        X_tes, T_tes = X[indi_tes], T[indi_tes]

        clf = CatBoostClassifier(verbose=False)

        clf.fit(X_tra, T_tra, verbose=False)

        ## Prediction
        pred_proba_tes = clf.predict_proba(X_tes)
        pred_cls_tes = np.argmax(pred_proba_tes, axis=1)

        pred_cls_tes[pred_cls_tes == 9] = 8  # overide 9 as 8 # fixme

        ##############################
        ## Manually adds objects to reporter to save
        ##############################
        ## Figure
        fig, ax = plt.subplots()
        ax.plot(np.arange(10))
        # reporter.add("manu_figs", fig)
        mreporter.add("manu_figs", fig, tgt="Test1")
        mreporter.add("manu_figs", fig, tgt="Test2")

        ## DataFrame
        df = pd.DataFrame(np.random.rand(5, 3))
        # reporter.add("manu_dfs", df)
        mreporter.add("manu_dfs", df, tgt="Test1")
        mreporter.add("manu_dfs", df, tgt="Test2")

        ## Scalar
        scalar = random.random()
        # reporter.add(
        #     "manu_scalars",
        #     scalar,
        # )
        mreporter.add("manu_scalars", scalar, tgt="Test1")
        mreporter.add("manu_scalars", scalar, tgt="Test2")

        ########################################
        ## Metrics
        ########################################
        mreporter.calc_metrics(
            T_tes,
            pred_cls_tes,
            pred_proba_tes,
            labels=labels,
            i_fold=i_fold,
            tgt="Test1",
        )
        mreporter.calc_metrics(
            T_tes,
            pred_cls_tes,
            pred_proba_tes,
            labels=labels,
            i_fold=i_fold,
            tgt="Test2",
        )

    # reporter.summarize(show=True)
    mreporter.summarize(show=True, tgt="Test1")
    mreporter.summarize(show=True, tgt="Test2")

    fake_fpaths = ["fake_file_1.txt", "fake_file_2.txt"]
    for ff in fake_fpaths:
        mngs.io.touch(ff)

    files_to_reproduce = [
        mngs.general.get_this_fpath(when_ipython="/dev/null"),
        *fake_fpaths,
    ]
    # reporter.save(files_to_reproduce=files_to_reproduce)
    mreporter.save(files_to_reproduce=files_to_reproduce, tgt="Test1")
    mreporter.save(files_to_reproduce=files_to_reproduce, tgt="Test2")

    confmat_plt_config = dict(
        figsize=(8, 8),
        # labelsize=8,
        # fontsize=6,
        # legendfontsize=6,
        figscale=2,
        tick_size=0.8,
        tick_width=0.2,
    )

    sci_notation_kwargs = dict(
        order=1,
        fformat="%1.0d",
        scilimits=(-3, 3),
        x=False,
        y=True,
    )  # "%3.1f"

    # sci_notation_kwargs = None
    # reporter.plot_and_save_conf_mats(
    #     plt,
    #     extend_ratio=1.0,
    #     confmat_plt_config=confmat_plt_config,
    #     sci_notation_kwargs=sci_notation_kwargs,
    # )

    mreporter.plot_and_save_conf_mats(
        plt,
        extend_ratio=1.0,
        confmat_plt_config=confmat_plt_config,
        sci_notation_kwargs=sci_notation_kwargs,
        tgt="Test1",
    )
    mreporter.plot_and_save_conf_mats(
        plt,
        extend_ratio=1.0,
        confmat_plt_config=confmat_plt_config,
        sci_notation_kwargs=sci_notation_kwargs,
        tgt="Test2",
    )

    ## EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-13 02:35:11 (ywatanabe)"


import torch
import torchaudio.transforms as T
from mngs.general import torch_fn


@torch_fn
def resample(x, src_fs, tgt_fs, t=None):
    xr = T.Resample(src_fs, tgt_fs, dtype=x.dtype).to(x.device)(x)
    if t is None:
        return xr
    if t is not None:
        tr = torch.linspace(t[0], t[-1], xr.shape[-1])
        return xr, tr


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs

    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # Parameters
    T_SEC = 1
    SIG_TYPE = "chirp"
    SRC_FS = 128
    TGT_FS_UP = 256
    TGT_FS_DOWN = 64
    FREQS_HZ = [10, 30, 100, 300]

    # Demo Signal
    xx, tt, fs = mngs.dsp.demo_sig(
        t_sec=T_SEC, fs=SRC_FS, freqs_hz=FREQS_HZ, sig_type=SIG_TYPE
    )

    # Resampling
    xd, td = mngs.dsp.resample(xx, fs, TGT_FS_DOWN, t=tt)
    xu, tu = mngs.dsp.resample(xx, fs, TGT_FS_UP, t=tt)

    # Plots
    i_batch, i_ch = 0, 0
    fig, axes = plt.subplots(nrows=3, sharex=True, sharey=True)
    axes[0].plot(tt, xx[i_batch, i_ch], label=f"Original ({SRC_FS} Hz)")
    axes[1].plot(
        td, xd[i_batch, i_ch], label=f"Down-sampled ({TGT_FS_DOWN} Hz)"
    )
    axes[2].plot(tu, xu[i_batch, i_ch], label=f"Up-sampled ({TGT_FS_UP} Hz)")
    for ax in axes:
        ax.legend(loc="upper left")

    axes[-1].set_xlabel("Time [s]")
    fig.supylabel("Amplitude [?V]")
    fig.suptitle("Resampling")
    mngs.io.save(fig, "traces.png")
    # plt.show()

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/dsp/_resample.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-04-23 11:02:45 (ywatanabe)"

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchsummary import summary
import mngs
import numpy as np


class ChannelGainChanger(nn.Module):
    def __init__(
        self,
        n_chs,
    ):
        super().__init__()
        self.n_chs = n_chs

    def forward(self, x):
        """x: [batch_size, n_chs, seq_len]"""
        if self.training:
            ch_gains = (
                torch.rand(self.n_chs).unsqueeze(0).unsqueeze(-1).to(x.device) + 0.5
            )
            ch_gains = F.softmax(ch_gains, dim=1)
            x *= ch_gains

        return x


if __name__ == "__main__":
    ## Demo data
    bs, n_chs, seq_len = 16, 360, 1000
    x = torch.rand(bs, n_chs, seq_len)

    cgc = ChGainChanger(n_chs)
    print(cgc(x).shape)  # [16, 19, 1000]

    # sb = SubjectBlock(n_chs=n_chs)
    # print(sb(x, s).shape) # [16, 270, 1000]

    # summary(sb, x, s)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-06-10 20:48:53 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/general/_mask_api_key.py


def mask_api(api_key, n=4):
    return f"{api_key[:n]}****{api_key[-n:]}"



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-06 07:16:38 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/general/_symlog.py

import numpy as np


def symlog(x, linthresh=1.0):
    """
    Apply a symmetric log transformation to the input data.

    Parameters
    ----------
    x : array-like
        Input data to be transformed.
    linthresh : float, optional
        Range within which the transformation is linear. Defaults to 1.0.

    Returns
    -------
    array-like
        Symmetrically transformed data.
    """
    sign_x = np.sign(x)
    abs_x = np.abs(x)
    return sign_x * (np.log1p(abs_x / linthresh))



import matplotlib
import numpy as np


class OOMFormatter(matplotlib.ticker.ScalarFormatter):
    def __init__(self, order=0, fformat="%1.1f", offset=True, mathText=True):
        self.order = order
        self.fformat = fformat
        matplotlib.ticker.ScalarFormatter.__init__(
            self, useOffset=offset, useMathText=mathText
        )

    def _set_order_of_magnitude(self):
        self.orderOfMagnitude = self.order

    def _set_format(self, vmin=None, vmax=None):
        self.format = self.fformat
        if self._useMathText:
            self.format = r"$\mathdefault{%s}$" % self.format


def sci_note(ax, fformat="%1.1f", x=False, y=False, scilimits=(-3, 3)):
    order_x = 0
    order_y = 0

    if x:
        order_x = np.floor(np.log10(np.max(np.abs(ax.get_xlim())) + 1e-5))
        ax.xaxis.set_major_formatter(
            OOMFormatter(order=int(order_x), fformat=fformat)
        )
        ax.ticklabel_format(axis="x", style="sci", scilimits=scilimits)
        ax.xaxis.labelpad = -22
        shift_x = (ax.get_xlim()[0] - ax.get_xlim()[1]) * 0.01
        ax.xaxis.get_offset_text().set_position((shift_x, 0))

    if y:
        order_y = np.floor(np.log10(np.max(np.abs(ax.get_ylim())) + 1e-5))
        ax.yaxis.set_major_formatter(
            OOMFormatter(order=int(order_y), fformat=fformat)
        )
        ax.ticklabel_format(axis="y", style="sci", scilimits=scilimits)
        ax.yaxis.labelpad = -20
        shift_y = (ax.get_ylim()[0] - ax.get_ylim()[1]) * 0.01
        ax.yaxis.get_offset_text().set_position((0, shift_y))

    return ax


# class OOMFormatter(matplotlib.ticker.ScalarFormatter):
#     def __init__(self, order=0, fformat="%1.1f", offset=True, mathText=True):
#         self.order = order
#         self.fformat = fformat
#         matplotlib.ticker.ScalarFormatter.__init__(
#             self, useOffset=offset, useMathText=mathText
#         )

#     def _set_order_of_magnitude(self):
#         self.orderOfMagnitude = self.order

#     def _set_format(self, vmin=None, vmax=None):
#         self.format = self.fformat
#         if self._useMathText:
#             self.format = r"$\mathdefault{%s}$" % self.format


# def sci_note(ax, fformat="%1.1f", x=False, y=False, scilimits=(-3, 3)):
#     order_x = 0
#     order_y = 0

#     if x:
#         order_x = np.floor(np.log10(np.max(np.abs(ax.get_xlim())) + 1e-5))
#         ax.xaxis.set_major_formatter(
#             OOMFormatter(order=int(order_x), fformat=fformat)
#         )
#         ax.ticklabel_format(axis="x", style="sci", scilimits=scilimits)

#     if y:
#         order_y = np.floor(np.log10(np.max(np.abs(ax.get_ylim()) + 1e-5)))
#         ax.yaxis.set_major_formatter(
#             OOMFormatter(order=int(order_y), fformat=fformat)
#         )
#         ax.ticklabel_format(axis="y", style="sci", scilimits=scilimits)

#     return ax


# #!/usr/bin/env python3


# import matplotlib


# class OOMFormatter(matplotlib.ticker.ScalarFormatter):
#     # https://stackoverflow.com/questions/42656139/set-scientific-notation-with-fixed-exponent-and-significant-digits-for-multiple
#     # def __init__(self, order=0, fformat="%1.1f", offset=True, mathText=True):
#     def __init__(self, order=0, fformat="%1.0d", offset=True, mathText=True):
#         self.oom = order
#         self.fformat = fformat
#         matplotlib.ticker.ScalarFormatter.__init__(
#             self, useOffset=offset, useMathText=mathText
#         )

#     def _set_order_of_magnitude(self):
#         self.orderOfMagnitude = self.oom

#     def _set_format(self, vmin=None, vmax=None):
#         self.format = self.fformat
#         if self._useMathText:
#             self.format = r"$\mathdefault{%s}$" % self.format


# def sci_note(
#     ax,
#     order,
#     fformat="%1.0d",
#     x=False,
#     y=False,
#     scilimits=(-3, 3),
# ):
#     """
#     Change the expression of the x- or y-axis to the scientific notation like *10^3
#     , where 3 is the first argument, order.

#     Example:
#         order = 4 # 10^4
#         ax = sci_note(
#                  ax,
#                  order,
#                  fformat="%1.0d",
#                  x=True,
#                  y=False,
#                  scilimits=(-3, 3),
#     """

#     if x == True:
#         ax.xaxis.set_major_formatter(
#             OOMFormatter(order=order, fformat=fformat)
#         )
#         ax.ticklabel_format(axis="x", style="sci", scilimits=scilimits)
#     if y == True:
#         ax.yaxis.set_major_formatter(
#             OOMFormatter(order=order, fformat=fformat)
#         )
#         ax.ticklabel_format(axis="y", style="sci", scilimits=scilimits)

#     return ax



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: 2024-05-12 21:02:21 (7)
# /sshx:ywatanabe@444:/home/ywatanabe/proj/mngs/src/mngs/general/_title2spath.py


def title2path(title):
    """
    Convert a title (string or dictionary) to a path-friendly string.

    Parameters
    ----------
    title : str or dict
        The input title to be converted.

    Returns
    -------
    str
        A path-friendly string derived from the input title.
    """
    if isinstance(title, dict):
        from mngs.general import dict2str

        title = dict2str(title)

    path = title

    patterns = [":", ";", "=", "[", "]"]
    for pattern in patterns:
        path = path.replace(pattern, "")

    path = path.replace("_-_", "-")
    path = path.replace(" ", "_")

    while "__" in path:
        path = path.replace("__", "_")

    return path.lower()


# def title2path(title):
#     if isinstance(title, dict):
#         title = dict2str(title)

#     path = title

#     # Comma patterns
#     patterns = [":", ";", "=", "[", "]"]
#     for pp in patterns:
#         path = path.replace(pp, "")

#     # Exceptions
#     path = path.replace("_-_", "-")
#     path = path.replace(" ", "_")

#     # Consective under scores
#     for _ in range(10):
#         path = path.replace("__", "_")

#     return path.lower()



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-29 22:10:06 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/general/data_processing/_to_rank.py

import torch
# from ..decorators._converters import
from mngs.gen import torch_fn


@torch_fn
def to_rank(tensor, method="average"):
    sorted_tensor, indices = torch.sort(tensor)
    ranks = torch.empty_like(tensor)
    ranks[indices] = (
        torch.arange(len(tensor), dtype=tensor.dtype, device=tensor.device) + 1
    )

    if method == "average":
        ranks = ranks.float()
        ties = torch.nonzero(sorted_tensor[1:] == sorted_tensor[:-1])
        for i in range(len(ties)):
            start = ties[i]
            end = start + 1
            while (
                end < len(sorted_tensor)
                and sorted_tensor[end] == sorted_tensor[start]
            ):
                end += 1
            ranks[indices[start:end]] = ranks[indices[start:end]].mean()

    return ranks



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-31 08:11:45 (ywatanabe)"

import torch


def apply_to(fn, x, dim):
    """
    Example:
    x = torch.randn(2, 3, 4)
    fn = sum
    apply_to(fn, x, 1).shape # (2, 1, 4)
    """
    if dim != -1:
        dims = list(range(x.dim()))
        dims[-1], dims[dim] = dims[dim], dims[-1]
        x = x.permute(*dims)

    # Flatten the tensor along the time dimension
    shape = x.shape
    x = x.reshape(-1, shape[-1])

    # Apply the function to each slice along the time dimension
    applied = torch.stack([fn(x_i) for x_i in torch.unbind(x, dim=0)], dim=0)

    # Reshape the tensor to its original shape (with the time dimension at the end)
    applied = applied.reshape(*shape[:-1], -1)

    # Permute back to the original dimension order if necessary
    if dim != -1:
        applied = applied.permute(*dims)

    return applied



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-20 19:42:38 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/io/_cache.py


import os
import pickle
import sys
from pathlib import Path


def cache(id, *args):
    """
    Store or fetch data using a pickle file.

    This function provides a simple caching mechanism for storing and retrieving
    Python objects. It uses pickle to serialize the data and stores it in a file
    with a unique identifier. If the data is already cached, it can be retrieved
    without recomputation.

    Parameters:
    -----------
    id : str
        A unique identifier for the cache file.
    *args : str
        Variable names to be cached or loaded.

    Returns:
    --------
    tuple
        A tuple of cached values corresponding to the input variable names.

    Raises:
    -------
    ValueError
        If the cache file is not found and not all variables are defined.

    Example:
    --------
    >>> import mngs
    >>> import numpy as np
    >>> 
    >>> # Variables to cache
    >>> var1 = "x"
    >>> var2 = 1
    >>> var3 = np.ones(10)
    >>> 
    >>> # Saving
    >>> var1, var2, var3 = mngs.io.cache("my_id", "var1", "var2", "var3")
    >>> print(var1, var2, var3)
    >>> 
    >>> # Loading when not all variables are defined and the id exists
    >>> del var1, var2, var3
    >>> var1, var2, var3 = mngs.io.cache("my_id", "var1", "var2", "var3")
    >>> print(var1, var2, var3)
    """
    cache_dir = Path.home() / ".cache" / "your_app_name"
    cache_dir.mkdir(parents=True, exist_ok=True)
    cache_file = cache_dir / f"{id}.pkl"

    does_cache_file_exist = cache_file.exists()

    # Get the caller's local variables
    caller_locals = sys._getframe(1).f_locals
    are_all_variables_defined = all(arg in caller_locals for arg in args)

    if are_all_variables_defined:
        # If all variables are defined, save them to cache and return as-is
        data_to_cache = {arg: caller_locals[arg] for arg in args}
        with cache_file.open("wb") as f:
            pickle.dump(data_to_cache, f)
        return tuple(data_to_cache.values())
    else:
        if does_cache_file_exist:
            # If cache exists, load and return the values
            with cache_file.open("rb") as f:
                loaded_data = pickle.load(f)
            return tuple(loaded_data[arg] for arg in args)
        else:
            raise ValueError(
                "Cache file not found and not all variables are defined."
            )


# Usage example
if __name__ == "__main__":
    import mngs
    import numpy as np

    # Variables to cache
    var1 = "x"
    var2 = 1
    var3 = np.ones(10)

    # Saving
    var1, var2, var3 = mngs.io.cache("my_id", "var1", "var2", "var3")
    print(var1, var2, var3)

    # Loading when not all variables are defined and the id exists
    del var1, var2, var3
    var1, var2, var3 = mngs.io.cache("my_id", "var1", "var2", "var3")
    print(var1, var2, var3)




#!/usr/bin/env python3


def shift(ax, dx=0, dy=0):
    """
    Adjusts the position of an Axes object within a Figure by specified offsets in centimeters.

    This function modifies the position of a given matplotlib.axes.Axes object by shifting it horizontally and vertically within its parent figure. The shift amounts are specified in centimeters, and the function converts these values into the figure's coordinate system to perform the adjustment.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to modify. This must be an instance of a Matplotlib Axes.
    - dx (float): The horizontal offset in centimeters. Positive values shift the Axes to the right, while negative values shift it to the left.
    - dy (float): The vertical offset in centimeters. Positive values shift the Axes up, while negative values shift it down.

    Returns:
    - matplotlib.axes.Axes: The modified Axes object with the adjusted position.
    """

    bbox = ax.get_position()

    # Convert centimeters to inches for consistency with matplotlib dimensions
    dx_in, dy_in = dx / 2.54, dy / 2.54

    # Calculate delta ratios relative to the figure size
    fig = ax.get_figure()
    fig_dx_in, fig_dy_in = fig.get_size_inches()
    dx_ratio, dy_ratio = dx_in / fig_dx_in, dy_in / fig_dy_in

    # Determine updated bbox position and optionally adjust dimensions
    left = bbox.x0 + dx_ratio
    bottom = bbox.y0 + dy_ratio
    width = bbox.width
    height = bbox.height

    # Main
    ax.set_position([left, bottom, width, height])

    return ax


# def adjust_axes_position_and_dimension(
#     ax, dx, dy, adjust_width_for_dx=False, adjust_height_for_dy=False
# ):

# def set_pos(ax, x_cm, y_cm, extend_x=False, extend_y=False):
#     """
#     Adjusts the position of an Axes object within a Figure by a specified offset in centimeters.

#     Parameters:
#     - ax (matplotlib.axes.Axes): The Axes object to modify.
#     - x_cm (float): The horizontal offset in centimeters to adjust the Axes position.
#     - y_cm (float): The vertical offset in centimeters to adjust the Axes position.
#     - extend_x (bool): If True, reduces the width of the Axes by the horizontal offset.
#     - extend_y (bool): If True, reduces the height of the Axes by the vertical offset.

#     Returns:
#     - ax (matplotlib.axes.Axes): The modified Axes object with the adjusted position.
#     """

#     bbox = ax.get_position()

#     # Inches
#     x_in, y_in = x_cm / 2.54, y_cm / 2.54

#     # Calculates delta ratios
#     fig = ax.get_figure()
#     fig_x_in, fig_y_in = fig.get_size_inches()
#     x_ratio, y_ratio = x_in / fig_x_in, y_in / fig_y_in

#     # Determines updated bbox position
#     left = bbox.x0 + x_ratio
#     bottom = bbox.y0 + y_ratio
#     width = bbox.width
#     height = bbox.height

#     if extend_x:
#         width -= x_ratio

#     if extend_y:
#         height -= y_ratio

#     ax.set_position([left, bottom, width, height])

#     return ax


# def set_pos(
#     fig,
#     ax,
#     x_cm,
#     y_cm,
#     dragh=False,
#     dragv=False,
# ):

#     bbox = ax.get_position()

#     ## Calculates delta ratios
#     fig_x_in, fig_y_in = fig.get_size_inches()

#     x_in = float(x_cm) / 2.54
#     y_in = float(y_cm) / 2.54

#     x_ratio = x_in / fig_x_in
#     y_ratio = y_in / fig_x_in

#     ## Determines updated bbox position
#     left = bbox.x0 + x_ratio
#     bottom = bbox.y0 + y_ratio
#     width = bbox.x1 - bbox.x0
#     height = bbox.y1 - bbox.y0

#     if dragh:
#         width -= x_ratio

#     if dragv:
#         height -= y_ratio

#     ax.set_pos(
#         [
#             left,
#             bottom,
#             width,
#             height,
#         ]
#     )

#     return ax



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-24 12:02:42 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/linalg/_distance.py


import scipy.spatial.distance as _distance
from mngs.gen import wrap, numpy_fn
import numpy as np


@numpy_fn
def euclidean_distance(uu, vv, axis=0):
    """
    Compute the Euclidean distance between two arrays along the specified axis.

    Parameters
    ----------
    uu : array_like
        First input array.
    vv : array_like
        Second input array.
    axis : int, optional
        Axis along which to compute the distance. Default is 0.

    Returns
    -------
    array_like
        Euclidean distance array along the specified axis.
    """
    uu, vv = np.atleast_1d(uu), np.atleast_1d(vv)

    if uu.shape[axis] != vv.shape[axis]:
        raise ValueError(f"Shape along axis {axis} must match")

    uu = np.moveaxis(uu, axis, 0)
    vv = np.moveaxis(vv, axis, 0)

    uu_tgt_shape = [uu.shape[0]] + list(uu.shape[1:]) + [1] * (vv.ndim - 1)
    vv_tgt_shape = [vv.shape[0]] + [1] * (uu.ndim - 1) + list(vv.shape[1:])

    uu_reshaped = uu.reshape(uu_tgt_shape)
    vv_reshaped = vv.reshape(vv_tgt_shape)

    diff = uu_reshaped - vv_reshaped
    euclidean_dist = np.sqrt(np.sum(diff**2, axis=axis))
    return euclidean_dist


@wrap
def cdist(*args, **kwargs):
    return _distance.cdist(*args, **kwargs)


edist = euclidean_distance

# Optionally, manually copy the original docstring
# euclidean_distance.__doc__ = _distance.euclidean.__doc__
cdist.__doc__ = _distance.cdist.__doc__



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-15 18:55:49"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

import sys

import matplotlib.pyplot as plt

# Imports
import mngs
from matplotlib.patches import Rectangle


# Functions
def rectangle(ax, xx, yy, ww, hh, **kwargs):
    ax.add_patch(Rectangle((xx, yy), ww, hh, **kwargs))
    return ax


# (YOUR AWESOME CODE)

if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/plt/ax/_rectangle.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-05-15 16:46:54 (ywatanabe)"

import torch
import torch.nn as nn
from torchsummary import summary

class ResNet1D(nn.Module):
    """
    A representative convolutional neural network for signal classification tasks.
    """

    def __init__(
        self,
            n_chs=19,
            n_out=10,
            n_blks=5
    ):
        super().__init__()

        # Parameters
        N_CHS = n_chs
        _N_FILTS_PER_CH = 4
        N_FILTS = N_CHS * _N_FILTS_PER_CH
        N_BLKS = n_blks

        # Convolutional layers
        self.res_conv_blk_layers = nn.Sequential(
            ResNetBasicBlock(N_CHS, N_FILTS),
            *[ResNetBasicBlock(N_FILTS, N_FILTS) for _ in range(N_BLKS - 1)],
        )

        # ## FC layer
        # self.fc = nn.Sequential(
        #     nn.Linear(N_FILTS, 64),
        #     nn.ReLU(),
        #     nn.Dropout(p=0.5),
        #     nn.Linear(64, 32),
        #     nn.ReLU(),
        #     nn.Dropout(p=0.5),
        #     nn.Linear(32, n_out),
        # )

    def forward(self, x):
        x = self.res_conv_blk_layers(x)
        # x = x.mean(axis=-1)
        # x = self.fc(x)
        return x


class ResNetBasicBlock(nn.Module):
    """The basic block of the ResNet1D model"""

    def __init__(self, in_chs, out_chs):
        super(ResNetBasicBlock, self).__init__()
        self.in_chs = in_chs
        self.out_chs = out_chs

        self.conv7 = self.conv_k(in_chs, out_chs, k=7, p=3)
        self.bn7 = nn.BatchNorm1d(out_chs)
        self.activation7 = nn.ReLU()

        self.conv5 = self.conv_k(out_chs, out_chs, k=5, p=2)
        self.bn5 = nn.BatchNorm1d(out_chs)
        self.activation5 = nn.ReLU()

        self.conv3 = self.conv_k(out_chs, out_chs, k=3, p=1)
        self.bn3 = nn.BatchNorm1d(out_chs)
        self.activation3 = nn.ReLU()

        self.expansion_conv = self.conv_k(in_chs, out_chs, k=1, p=0)

        self.bn = nn.BatchNorm1d(out_chs)
        self.activation = nn.ReLU()

    @staticmethod
    def conv_k(in_chs, out_chs, k=1, s=1, p=1):
        """Build size k kernel's convolution layer with padding"""
        return nn.Conv1d(
            in_chs, out_chs, kernel_size=k, stride=s, padding=p, bias=False
        )

    def forward(self, x):
        residual = x

        x = self.conv7(x)
        x = self.bn7(x)
        x = self.activation7(x)

        x = self.conv5(x)
        x = self.bn5(x)
        x = self.activation5(x)

        x = self.conv3(x)
        x = self.bn3(x)
        x = self.activation3(x)

        if self.in_chs != self.out_chs:
            residual = self.expansion_conv(residual)
        residual = self.bn(residual)

        x = x + residual
        x = self.activation(x)

        return x
    
if __name__ == "__main__":
    import sys

    sys.path.append("./DEAP/")
    import utils

    # Demo data
    bs, n_chs, seq_len = 16, 32, 8064
    Xb = torch.rand(bs, n_chs, seq_len)

    model = ResNet1D(
            n_chs=n_chs,
            n_out=4,
    )  # utils.load_yaml("./config/global.yaml")["EMOTIONS"]
    y = model(Xb)  # 16,4
    summary(model, Xb)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 12:59:28 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/ai/_gen_ai/_PRICING.py


from .PARAMS import MODELS


def calc_cost(model, input_tokens, output_tokens):
    import pandas as pd
    MODELs = pd.DataFrame(MODELS)
    indi = MODELS["name"] == model
    costs = MODELS[["input_cost", "output_cost"]][indi]
    cost = (
        input_tokens * costs["input_cost"]
        + output_tokens * costs["output_cost"]
    ) / 1_000_000
    return cost.iloc[0]



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-29 14:34:11 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_Llama.py

"""
This script implements the Llama model using the BaseGenAI structure.
"""

"""
Imports
"""
import os
import sys
from typing import List, Optional

import matplotlib.pyplot as plt
import mngs

try:
    from llama import Dialog
    from llama import Llama as _Llama
except:
    pass

from ._BaseGenAI import BaseGenAI

"""
Functions & Classes
"""


def print_envs():
    settings = {
        "MASTER_ADDR": os.getenv("MASTER_ADDR", "localhost"),
        "MASTER_PORT": os.getenv("MASTER_PORT", "12355"),
        "WORLD_SIZE": os.getenv("WORLD_SIZE", "1"),
        "RANK": os.getenv("RANK", "0"),
    }

    print("Environment Variable Settings:")
    for key, value in settings.items():
        print(f"{key}: {value}")
    print()


class Llama(BaseGenAI):
    def __init__(
        self,
        ckpt_dir: str = "",
        tokenizer_path: str = "",
        system_setting: str = "",
        model: str = "Meta-Llama-3-8B",
        max_seq_len: int = 512,
        max_batch_size: int = 4,
        max_gen_len: Optional[int] = None,
        stream: bool = False,
        seed: Optional[int] = None,
        n_keep: int = 1,
        temperature: float = 1.0,
        provider="Llama",
        chat_history=None,
        **kwargs,
    ):

        # Configure environment variables
        os.environ["MASTER_ADDR"] = os.getenv("MASTER_ADDR", "localhost")
        os.environ["MASTER_PORT"] = os.getenv("MASTER_PORT", "12355")
        os.environ["WORLD_SIZE"] = os.getenv("WORLD_SIZE", "1")
        os.environ["RANK"] = os.getenv("RANK", "0")
        print_envs()

        self.ckpt_dir = (
            ckpt_dir if ckpt_dir else f"Meta-{model}/"
        )  # consolidated.00.pth
        self.tokenizer_path = (
            tokenizer_path
            if tokenizer_path
            else f"./Meta-{model}/tokenizer.model"
        )
        self.max_seq_len = max_seq_len
        self.max_batch_size = max_batch_size
        self.max_gen_len = max_gen_len

        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key="",  # Llama3 doesn't use an API key
            stream=stream,
            seed=seed,
            n_keep=n_keep,
            temperature=temperature,
            chat_history=chat_history,
        )

    def __str__(self):
        return "Llama"

    def _init_client(self):
        generator = _Llama.build(
            ckpt_dir=self.ckpt_dir,
            tokenizer_path=self.tokenizer_path,
            max_seq_len=self.max_seq_len,
            max_batch_size=self.max_batch_size,
        )
        return generator

    def _api_call_static(self):
        dialogs: List[Dialog] = [self.history]
        results = self.client.chat_completion(
            dialogs,
            max_gen_len=self.max_gen_len,
            temperature=self.temperature,
            top_p=0.9,
        )
        out_text = results[0]["generation"]["content"]
        return out_text

    def _api_call_stream(self):
        # Llama3 doesn't have built-in streaming, so we'll simulate it
        full_response = self._api_call_static()
        for char in full_response:
            yield char

    # def _get_available_models(self):
    #     # Llama3 doesn't have a list of available models, so we'll return a placeholder
    #     return ["llama3"]

    def verify_model(self):
        # Llama3 doesn't require model verification, so we'll skip it
        pass


def main():
    m = Llama(
        ckpt_dir="/path/to/checkpoint",
        tokenizer_path="/path/to/tokenizer",
        system_setting="You are a helpful assistant.",
        max_seq_len=512,
        max_batch_size=4,
        stream=True,
        temperature=0.7,
    )
    m("Hi")
    pass


if __name__ == "__main__":
    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-04 17:24:37 (ywatanabe)"

from ._add_marginal_ax import add_marginal_ax
from ._circular_hist import circular_hist
from ._ecdf import ecdf
from ._extend import extend
from ._fillv import fillv
from ._force_aspect import force_aspect
from ._format_label import format_label
from ._hide_spines import hide_spines
from ._imshow2d import imshow2d
from ._joyplot import joyplot
from ._map_ticks import map_ticks
from ._panel import panel
from ._plot_ import fill_between, plot_, plot_with_ci
from ._raster_plot import raster_plot
from ._rectangle import rectangle
from ._sci_note import sci_note
from ._set_n_ticks import set_n_ticks
from ._set_size import set_size
from ._set_supxyt import set_supxyt
from ._set_ticks import set_ticks
from ._set_xyt import set_xyt
from ._share_axes import (get_global_xlim, get_global_ylim, set_xlims,
                          set_ylims, sharex, sharexy, sharey)
# from ._set_pos import set_pos
from ._shift import shift
from ._rotate_labels import rotate_labels
from ._half_violin import half_violin



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-04-23 11:02:34 (ywatanabe)"

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchsummary import summary
import mngs
import numpy as np
import julius

# BANDS_LIM_HZ_DICT = {
#     "delta": [0.5, 4],
#     "theta": [4, 8],
#     "lalpha": [8, 10],
#     "halpha": [10, 13],
#     "beta": [13, 32],
#     "gamma": [32, 75],
# }


# class FreqDropout(nn.Module):
#     def __init__(self, n_bands, samp_rate, dropout_ratio=0.5):
#         super().__init__()
#         self.dropout = nn.Dropout(p=0.5)
#         self.n_bands = n_bands
#         self.samp_rate = samp_rate
#         # self.
#         self.register_buffer("ones", torch.ones(self.n_bands))

#     def forward(self, x):
#         """x: [batch_size, n_chs, seq_len]"""
#         x = julius.bands.split_bands(x, self.samp_rate, n_bands=self.n_bands)

#         gains_orig = x.reshape(len(x), -1).abs().sum(axis=-1)
#         sum_gains_orig = gains_orig.sum()

#         # use_freqs = self.dropout(torch.ones(self.n_bands)).bool().long()
#         use_freqs = self.dropout(self.ones) / 2 # .bool().long()        

#         gains = gains_orig * use_freqs
#         sum_gains = gains.sum()
#         gain_ratio = sum_gains / sum_gains_orig


#         x *= use_freqs.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
#         x /= gain_ratio
#         x = x.sum(axis=0)
        
#         return x

class FreqGainChanger(nn.Module):
    def __init__(self, n_bands, samp_rate, dropout_ratio=0.5):
        super().__init__()
        self.dropout = nn.Dropout(p=0.5)
        self.n_bands = n_bands
        self.samp_rate = samp_rate
        # self.register_buffer("ones", torch.ones(self.n_bands))

    def forward(self, x):
        """x: [batch_size, n_chs, seq_len]"""
        if self.training:
            x = julius.bands.split_bands(x, self.samp_rate, n_bands=self.n_bands)
            freq_gains = torch.rand(self.n_bands).unsqueeze(-1).unsqueeze(-1).unsqueeze(-1).to(x.device) + .5
            freq_gains = F.softmax(freq_gains, dim=0)
            x = (x*freq_gains).sum(axis=0)

        return x
        # import ipdb; ipdb.set_trace()

        # gains_orig = x.reshape(len(x), -1).abs().sum(axis=-1)
        # sum_gains_orig = gains_orig.sum()

        # # use_freqs = self.dropout(torch.ones(self.n_bands)).bool().long()
        # use_freqs = self.dropout(self.ones) / 2 # .bool().long()        

        # gains = gains_orig * use_freqs
        # sum_gains = gains.sum()
        # gain_ratio = sum_gains / sum_gains_orig


        # x *= use_freqs.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        # x /= gain_ratio
        # x = x.sum(axis=0)
        
        # return x



    

if __name__ == "__main__":
    # Parameters
    N_BANDS = 10
    SAMP_RATE = 1000
    BS, N_CHS, SEQ_LEN = 16, 360, 1000

    # Demo data
    x = torch.rand(BS, N_CHS, SEQ_LEN).cuda()

    # Feedforward
    fgc = FreqGainChanger(N_BANDS, SAMP_RATE).cuda()
    # fd.eval()
    y = fgc(x)
    y.sum().backward()

    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-12 09:53:15 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/ax/_raster_plot.py

"""This script provides a functionality of raster plotting"""

import sys
from bisect import bisect_left

import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd


def raster_plot(ax, positions, time=None, **kwargs):
    """
    Create a raster plot using eventplot and return the plot along with a DataFrame.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes on which to draw the raster plot.
    positions : list of lists or array-like
        Position of events for each channel. Each list corresponds to events of one channel.
    time : array-like, optional
        The time indices for the events. If None, time will be generated based on event positions.
    **kwargs : dict
        Additional keyword arguments to be passed to the eventplot function.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes with the raster plot.
    df : pandas.DataFrame
        A DataFrame where rows correspond to time indices and columns correspond to channels.
        Each cell contains the channel index for events at specific time indices.

    Example
    -------
    positions = [[10, 50, 90], [20, 60, 100], [30, 70, 110]]
    fig, ax = plt.subplots()
    ax, df = raster_plot(ax, positions)
    plt.show()
    """

    def ensure_list(positions):
        return [
            [pos] if isinstance(pos, (int, float)) else pos
            for pos in positions
        ]

    def positions_to_df(positions, time):
        if time is None:
            time = np.linspace(
                0, np.max([np.max(pos) for pos in positions]), 1000
            )

        digi = np.full((len(positions), len(time)), np.nan, dtype=float)

        for channel_index, channel_positions in enumerate(positions):
            for position in channel_positions:
                insert_index = bisect_left(time, position)
                if insert_index == len(time):
                    insert_index -= 1
                digi[channel_index, insert_index] = channel_index

        return pd.DataFrame(digi.T, index=time)

    positions = ensure_list(positions)
    df = positions_to_df(positions, time)

    ax.eventplot(positions, orientation="horizontal", **kwargs)
    return ax, df


def test():
    positions = [
        [10, 50, 90],
        [20, 60, 100],
        [30, 70, 110],
        [40, 80, 120],
    ]
    fig, ax = mngs.plt.subplots()
    ax, df = raster_plot(ax, positions)


if __name__ == "__main__":
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    test()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-13 19:05:01 (ywatanabe)"

try:
    from ._apply_to import apply_to
except ImportError as e:
    warnings.warn(f"Warning: Failed to import apply_to from ._apply_to.")

# from ._apply_to import apply_to



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 18:40:14 (ywatanabe)"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

# try:
#     from ._find import find_dir, find_file, find_git_root
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._find.")

# try:
#     from ._path import file_size, spath, split, this_path, getsize
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._path.")

# try:
#     from ._version import find_latest, increment_version
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import from ._version.")

from ._find import find_dir, find_file, find_git_root
from ._path import file_size, spath, split, this_path
from ._version import find_latest, increment_version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 18:39:23 (ywatanabe)"
# mngs_repo/src/mngs/linalg/__init__.py

from ._distance import cdist, euclidean_distance, edist
from ._geometric_median import geometric_median
from ._misc import cosine, nannorm, rebase_a_vec, three_line_lengths_to_coords

# # Import distance-related functions
# cdist, euclidean_distance, edist = try_import(
#     "._distance", ["cdist", "euclidean_distance", "edist"]
# )

# # Import geometric median
# geometric_median, = try_import(
#     "._geometric_median", ["geometric_median"]
# )

# # Import miscellaneous functions
# cosine, nannorm, rebase_a_vec, three_line_lengths_to_coords = try_import(
#     "._misc", ["cosine", "nannorm", "rebase_a_vec", "three_line_lengths_to_coords"]
# )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-05 23:04:16 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/pd/_melt_cols.py


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-05 23:03:39 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/pd/_melt_cols.py

from typing import List, Optional
import pandas as pd

def melt_cols(df: pd.DataFrame, cols: List[str], id_columns: Optional[List[str]] = None) -> pd.DataFrame:
    """
    Melt specified columns while preserving links to other data in a DataFrame.

    Example
    -------
    >>> data = pd.DataFrame({
    ...     'id': [1, 2, 3],
    ...     'name': ['Alice', 'Bob', 'Charlie'],
    ...     'score_1': [85, 90, 78],
    ...     'score_2': [92, 88, 95]
    ... })
    >>> melted = melt_cols(data, cols=['score_1', 'score_2'])
    >>> print(melted)
       id     name variable  value
    0   1    Alice  score_1     85
    1   2      Bob  score_1     90
    2   3  Charlie  score_1     78
    3   1    Alice  score_2     92
    4   2      Bob  score_2     88
    5   3  Charlie  score_2     95

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame
    cols : List[str]
        Columns to be melted
    id_columns : Optional[List[str]], default None
        Columns to preserve as identifiers. If None, all columns not in 'cols' are used.

    Returns
    -------
    pd.DataFrame
        Melted DataFrame with preserved identifier columns

    Raises
    ------
    ValueError
        If cols are not present in the DataFrame
    """
    missing_melt = set(cols) - set(df.columns)
    if missing_melt:
        raise ValueError(f"Columns not found in DataFrame: {missing_melt}")

    id_columns = id_columns or [col for col in df.columns if col not in cols]

    df_copy = df.reset_index(drop=True)
    df_copy['global_index'] = df_copy.index

    melted_df = df_copy[cols + ['global_index']].melt(id_vars=['global_index'])
    formatted_df = melted_df.merge(df_copy[id_columns + ['global_index']], on='global_index')

    return formatted_df.drop('global_index', axis=1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-13 18:53:04 (ywatanabe)"
# /home/yusukew/proj/mngs_repo/src/mngs/general/_print_config.py

"""
1. Functionality:
   - Prints configuration values from YAML files
2. Input:
   - Configuration key (dot-separated for nested structures)
3. Output:
   - Corresponding configuration value
4. Prerequisites:
   - mngs package with load_configs function

Example:
    python _print_config.py PATH.TITAN.MAT
"""


import sys
import os
import argparse
from pprint import pprint
import sys

def print_config(key):

    import mngs

    CONFIG = mngs.io.load_configs()

    if key is None:
        print("Available configurations:")
        pprint(CONFIG)
        return

    try:
        keys = key.split(".")
        value = CONFIG
        for k in keys:

            if isinstance(value, (dict, mngs.gen.utils._DotDict.DotDict)):
                value = value.get(k)

            elif isinstance(value, list):
                try:
                    value = value[int(k)]
                except (ValueError, IndexError):
                    value = None

            elif isinstance(value, str):
                break

            else:
                value = None

            if value is None:
                break

        print(value)

    except Exception as e:
        print(f"Error: {e}")
        print("Available configurations:")
        pprint(value)

def print_config_main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(description="Print configuration values")
    parser.add_argument("key", nargs='?', default=None, help="Configuration key (dot-separated for nested structures)")
    parsed_args = parser.parse_args(args)
    print_config(parsed_args.key)

if __name__ == "__main__":
    print_config_main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-02 23:32:31 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/tex/_to_vec.py

def to_vec(v_str):
    r"""
    Convert a string to LaTeX vector notation.

    Example
    -------
    vector = to_vec("AB")
    print(vector)  # Outputs: \overrightarrow{\mathrm{AB}}

    Parameters
    ----------
    vector_string : str
        String representation of the vector

    Returns
    -------
    str
        LaTeX representation of the vector
    """
    return f"\\overrightarrow{{\\mathrm{{{v_str}}}}}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 18:40:05 (ywatanabe)"

from ._mv import mv
# try:
#     from ._mv import mv
# except ImportError as e:
#     warnings.warn(f"Warning: Failed to import mv from ._mv.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-24 14:50:41 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/plt/ax/_rotate_labels.py

"""This script does XYZ."""

"""Imports"""
def rotate_labels(ax, x=45, y=45, x_ha='center', y_ha='center'):
    """
    Rotate x and y axis labels of a matplotlib Axes object.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The Axes object to modify.
    x : float, optional
        Rotation angle for x-axis labels in degrees. Default is 45.
    y : float, optional
        Rotation angle for y-axis labels in degrees. Default is 45.
    x_ha : str, optional
        Horizontal alignment for x-axis labels. Default is 'center'.
    y_ha : str, optional
        Horizontal alignment for y-axis labels. Default is 'center'.

    Returns
    -------
    matplotlib.axes.Axes
        The modified Axes object.

    Example
    -------
    fig, ax = plt.subplots()
    ax.plot([1, 2, 3], [1, 2, 3])
    rotate_labels(ax)
    plt.show()
    """
    ax.set_xticklabels(ax.get_xticklabels(), rotation=x, ha=x_ha)
    ax.set_yticklabels(ax.get_yticklabels(), rotation=y, ha=y_ha)
    return ax



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-24 18:40:30 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/pd/__init__.py

import os
import importlib
import inspect
import warnings

# Get the current directory
current_dir = os.path.dirname(__file__)

# Iterate through all Python files in the current directory
for filename in os.listdir(current_dir):
    if filename.endswith(".py") and not filename.startswith("__"):
        module_name = filename[:-3]
        # module = importlib.import_module(f".{module_name}", package=__name__)

        # # Import only functions and classes from the module
        # for name, obj in inspect.getmembers(module):
        #     if inspect.isfunction(obj) or inspect.isclass(obj):
        #         if not name.startswith("_"):
        #             globals()[name] = obj
        try:
            module = importlib.import_module(f".{module_name}", package=__name__)

            # Import only functions and classes from the module
            for name, obj in inspect.getmembers(module):
                if inspect.isfunction(obj) or inspect.isclass(obj):
                    if not name.startswith("_"):
                        globals()[name] = obj
        except ImportError as e:
            warnings.warn(f"Warning: Failed to import {module_name}.")

# Clean up temporary variables
del (
    os,
    importlib,
    inspect,
    current_dir,
    filename,
    module_name,
    module,
    name,
    obj,
)

# from ._merge_columns import merge_cols, merge_columns
# from ._misc import find_indi  # col_to_last,; col_to_top,; merge_columns,
# from ._misc import force_df, ignore_SettingWithCopyWarning, slice
# from ._mv import mv, mv_to_first, mv_to_last
# from ._sort import sort



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-06 11:09:07 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/stats/_find_pval_col.py

"""
Functionality:
    - Identifies column name(s) in a DataFrame or keys in other data structures that correspond to p-values
Input:
    - pandas DataFrame, numpy array, list, or dict
Output:
    - String or list of strings representing the identified p-value column name(s) or key(s), or None if not found
Prerequisites:
    - pandas, numpy libraries
"""

import re
from typing import Optional, Union, List, Dict
import pandas as pd
import numpy as np


def find_pval(
    data: Union[pd.DataFrame, np.ndarray, List, Dict], multiple: bool = True
) -> Union[Optional[str], List[str]]:
    """
    Find p-value column name(s) or key(s) in various data structures.

    Example:
    --------
    >>> df = pd.DataFrame({'p_value': [0.05, 0.01], 'pval': [0.1, 0.001], 'other': [1, 2]})
    >>> find_pval(df)
    ['p_value', 'pval']
    >>> find_pval(df, multiple=False)
    'p_value'

    Parameters:
    -----------
    data : Union[pd.DataFrame, np.ndarray, List, Dict]
        Data structure to search for p-value column or key
    multiple : bool, optional
        If True, return all matches; if False, return only the first match (default is True)

    Returns:
    --------
    Union[Optional[str], List[str]]
        Name(s) of the column(s) or key(s) that match p-value patterns, or None if not found
    """
    if isinstance(data, pd.DataFrame):
        return find_pval_col(data, multiple)
    elif isinstance(data, (np.ndarray, list, dict)):
        return _find_pval(data, multiple)
    else:
        raise ValueError(
            "Input must be a pandas DataFrame, numpy array, list, or dict"
        )


def _find_pval(
    data: Union[np.ndarray, List, Dict], multiple: bool
) -> Union[Optional[str], List[str]]:
    pattern = re.compile(r"p[-_]?val(ue)?(?!.*stars)", re.IGNORECASE)
    matches = []

    if isinstance(data, dict):
        matches = [key for key in data.keys() if pattern.search(str(key))]
    elif (
        isinstance(data, (np.ndarray, list))
        and len(data) > 0
        and isinstance(data[0], dict)
    ):
        matches = [key for key in data[0].keys() if pattern.search(str(key))]

    return matches if multiple else (matches[0] if matches else None)


def find_pval_col(
    df: pd.DataFrame, multiple: bool = False
) -> Union[Optional[str], List[str]]:
    """
    Find p-value column name(s) in a DataFrame.

    Example:
    --------
    >>> df = pd.DataFrame({'p_value': [0.05, 0.01], 'pval': [0.1, 0.001], 'other': [1, 2]})
    >>> find_pval_col(df)
    ['p_value', 'pval']
    >>> find_pval_col(df, multiple=False)
    'p_value'

    Parameters:
    -----------
    df : pd.DataFrame
        DataFrame to search for p-value column(s)
    multiple : bool, optional
        If True, return all matches; if False, return only the first match (default is False)

    Returns:
    --------
    Union[Optional[str], List[str]]
        Name(s) of the column(s) that match p-value patterns, or None if not found
    """
    pattern = re.compile(r"p[-_]?val(ue)?(?!.*stars)", re.IGNORECASE)
    matches = [col for col in df.columns if pattern.search(str(col))]

    return matches if multiple else (matches[0] if matches else None)



#!/usr/bin/env python

import os
from setuptools import find_packages, setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


setup(
    name='ranger',
    version='0.1.dev0',
    packages=find_packages(
        exclude=['tests', '*.tests', '*.tests.*', 'tests.*']
    ),
    package_dir={'ranger': os.path.join('.', 'ranger')},
    description='Ranger - a synergistic optimizer using RAdam '
                '(Rectified Adam) and LookAhead in one codebase ',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Less Wright',
    license='Apache',
    install_requires=['torch']
)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-07 21:19:31 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/plt/_subplots/AxisWrapper.py

from collections import OrderedDict
from contextlib import contextmanager
from functools import wraps

import mngs
import numpy as np
import pandas as pd
import seaborn as sns
from mngs.gen import not_implemented
from scipy.stats import gaussian_kde

from ._to_sigma import to_sigma as _to_sigma


def sns_copy_doc(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        return func(self, *args, **kwargs)

    wrapper.__doc__ = getattr(sns, func.__name__.split("sns_")[-1]).__doc__
    return wrapper


class AxisWrapper:
    """
    A wrapper class for a Matplotlib axis that collects plotting data.
    """

    def __init__(self, fig, axis, track):
        """
        Initialize the AxisWrapper with a given axis and history reference.
        """
        self.fig = fig
        self.axis = axis
        self._ax_history = OrderedDict()
        self.track = track
        self.id = 0

    def get_figure(
        self,
    ):
        return self.fig

    def __getattr__(self, attr):
        if hasattr(self.axis, attr):
            original_attr = getattr(self.axis, attr)

            if callable(original_attr):

                @wraps(original_attr)
                def wrapper(*args, track=None, id=None, **kwargs):
                    results = original_attr(*args, **kwargs)
                    self._track(track, id, attr, args, kwargs)
                    return results

                return wrapper
            else:
                return original_attr
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{attr}'"
        )

    ################################################################################
    ## Tracking
    ################################################################################
    def _track(self, track, id, method_name, args, kwargs):
        if track is None:
            track = self.track
        if track:
            id = id if id is not None else self.id
            self.id += 1
            self._ax_history[id] = (id, method_name, args, kwargs)

    @contextmanager
    def _no_tracking(self):
        """Context manager to temporarily disable tracking."""
        original_track = self.track
        self.track = False
        try:
            yield
        finally:
            self.track = original_track

    @property
    def history(self):
        return {k: self._ax_history[k] for k in self._ax_history}

    @property
    def flat(self):
        if isinstance(self.axis, list):
            return self.axis
        else:
            return [self.axis]

    def reset_history(self):
        self._ax_history = {}

    def to_sigma(self):
        """
        Export tracked plotting data to a DataFrame in SigmaPlot format.
        """
        df = _to_sigma(self.history)

        return df if df is not None else pd.DataFrame()

    ################################################################################
    ## Plotting methods (other than matplotlib's original methods)
    ################################################################################
    def imshow2d(
        self,
        arr_2d,
        cbar=True,
        cbar_label=None,
        cbar_shrink=1.0,
        cbar_fraction=0.046,
        cbar_pad=0.04,
        cmap="viridis",
        aspect="auto",
        vmin=None,
        vmax=None,
        track=True,
        id=None,
        xyz=False,
        **kwargs,
    ):
        # Method name
        method_name = "imshow2d"

        # Plotting
        with self._no_tracking():
            self.axis = mngs.plt.ax.imshow2d(
                self.axis,
                arr_2d,
                cbar=cbar,
                cbar_label=cbar_label,
                cbar_shrink=cbar_shrink,
                cbar_fraction=cbar_fraction,
                cbar_pad=cbar_pad,
                cmap=cmap,
                aspect=aspect,
                vmin=vmin,
                vmax=vmax,
            )

        # Tracking
        out = pd.DataFrame(arr_2d)
        if xyz:
            out = mngs.pd.to_xyz(out)

        self._track(track, id, method_name, out, None)

    def conf_mat(
        self,
        data,
        x_labels=None,
        y_labels=None,
        title="Confusion Matrix",
        cmap="Blues",
        cbar=True,
        cbar_kw={},
        label_rotation_xy=(15, 15),
        x_extend_ratio=1.0,
        y_extend_ratio=1.0,
        bacc=False,
        track=True,
        id=None,
        **kwargs,
    ):
        # Method name
        method_name = "conf_mat"

        # Plotting
        with self._no_tracking():
            out = mngs.plt.ax.conf_mat(
                self.axis,
                data,
                x_labels=x_labels,
                y_labels=y_labels,
                title=title,
                cmap=cmap,
                cbar=cbar,
                cbar_kw=cbar_kw,
                label_rotation_xy=label_rotation_xy,
                x_extend_ratio=x_extend_ratio,
                y_extend_ratio=y_extend_ratio,
                bacc=bacc,
                track=track,
                id=id,
                **kwargs,
            )
            bacc_val = None

            if bacc:
                self.axis, bacc_val = out
            else:
                self.axis = out

        # Tracking
        out = data, bacc_val
        self._track(track, id, method_name, out, None)

    def rectangle(self, xx, yy, ww, hh, track=True, id=None, **kwargs):
        # Method name
        method_name = "rectangle"

        # Plotting
        with self._no_tracking():
            self.axis = mngs.plt.ax.rectangle(
                self.axis, xx, yy, ww, hh, **kwargs
            )

        # Tracking
        out = None
        self._track(track, id, method_name, out, None)

    def plot_(
        self,
        xx=None,
        yy=None,
        mean=None,
        median=None,
        std=None,
        ci=None,
        iqr=None,
        n=None,
        alpha=0.3,
        track=True,
        id=None,
        **kwargs,
    ):
        # Method
        method_name = "plot_"

        # Plotting
        with self._no_tracking():
            self.axis, df = mngs.plt.ax.plot_(
                self.axis,
                xx=xx,
                yy=yy,
                mean=mean,
                median=median,
                std=std,
                ci=ci,
                iqr=iqr,
                n=n,
                alpha=alpha,
                **kwargs,
            )

        # Tracking
        out = df
        self._track(track, id, method_name, out, None)

    def fillv(
        self,
        starts,
        ends,
        color="red",
        alpha=0.2,
        track=True,
        id=None,
        **kwargs,
    ):
        # Method name
        method_name = "fillv"

        self.axis = mngs.plt.ax.fillv(
            self.axis, starts, ends, color=color, alpha=alpha
        )

        # Tracking
        out = (starts, ends)
        self._track(track, id, method_name, out, None)

    def boxplot_(self, data, track=True, id=None, **kwargs):
        # Method name
        method_name = "boxplot_"

        # Deep Copy
        _data = data.copy()

        # # NaN Handling
        # data = np.hstack(data)
        # data = data[~np.isnan(data)]

        n = len(data)

        if kwargs.get("label"):
            kwargs["label"] = kwargs["label"] + f" (n={n})"

        # Plotting
        with self._no_tracking():
            self.axis.boxplot(data, **kwargs)

        out = pd.DataFrame(
            {
                "data": _data,
                "n": [n for _ in range(len(data))],
            }
        )

        # Tracking
        self._track(track, id, method_name, out, None)

    def raster(self, positions, time=None, track=True, id=None, **kwargs):
        # Method name
        method_name = "raster"

        # Plotting
        with self._no_tracking():
            self.axis, df = mngs.plt.ax.raster_plot(
                self.axis, positions, time=time, **kwargs
            )
        if id is not None:
            df.columns = [f"{id}_{method_name}_{col}" for col in df.columns]
        out = df

        # Tracking
        self._track(track, id, method_name, out, None)

    def kde(self, data, track=True, id=None, xlim=None, **kwargs):
        # Method name
        method_name = "kde"

        # NaN Handling
        n = (~np.isnan(data)).sum()

        if kwargs.get("label"):
            kwargs["label"] = kwargs["label"] + f" (n={n})"

        # x axis
        xlim = xlim if xlim is not None else (data.min(), data.max())
        xs = np.linspace(*xlim, int(1e3))

        # KDE
        density = gaussian_kde(data)(xs)
        density /= density.sum()

        # Cumulative
        if kwargs.get("cumulative"):
            density = np.cumsum(density)

        # Plotting
        with self._no_tracking():
            if kwargs.get("fill"):
                self.axis.fill_between(xs, density, **kwargs)
            else:
                self.plot_(xx=xs, yy=density, label=kwargs.get("label"))

        # Tracking
        out = pd.DataFrame(
            {"x": xs, "kde": density, "n": [len(data) for _ in range(len(xs))]}
        )
        self._track(track, id, method_name, out, None)

    def ecdf(self, data, track=True, id=None, **kwargs):
        # Method name
        method_name = "ecdf"

        # Plotting
        with self._no_tracking():
            self.axis, df = mngs.plt.ax.ecdf(self.axis, data, **kwargs)
        out = df

        # Tracking
        self._track(track, id, method_name, out, None)

    def joyplot(
        self,
        data,
        track=True,
        id=None,
        **kwargs,
    ):
        # Method name
        method_name = "joyplot"

        # Plotting
        with self._no_tracking():
            self.axis = mngs.plt.ax.joyplot(self.axis, data, **kwargs)

        # Tracking
        out = data
        self._track(track, id, method_name, out, None)

    ################################################################################
    ## Seaborn-wrappers
    ################################################################################
    def _sns_base(
        self, method_name, *args, track=True, track_obj=None, id=None, **kwargs
    ):
        sns_method_name = method_name.split("sns_")[-1]

        with self._no_tracking():
            sns_plot_fn = getattr(sns, sns_method_name)

            if kwargs.get("hue_colors"):
                kwargs = mngs.gen.alternate_kwarg(
                    kwargs, primary_key="palette", alternate_key="hue_colors"
                )

            self.axis = sns_plot_fn(ax=self.axis, *args, **kwargs)

        # Track the plot if required
        track_obj = track_obj if track_obj is not None else args
        self._track(track, id, method_name, track_obj, kwargs)

    def _sns_base_xyhue(
        self, method_name, *args, track=True, id=None, **kwargs
    ):
        """Formats data passed to sns functions with (data=data, x=x, y=y) keyword arguments"""
        df = kwargs.get("data")
        x, y, hue = kwargs.get("x"), kwargs.get("y"), kwargs.get("hue")

        track_obj = (
            self._sns_prepare_xyhue(df, x, y, hue) if df is not None else None
        )
        self._sns_base(
            method_name,
            *args,
            track=track,
            track_obj=track_obj,
            id=id,
            **kwargs,
        )

    def _sns_prepare_xyhue(
        self, data=None, x=None, y=None, hue=None, **kwargs
    ):
        """Returns obj to track"""
        data = data.reset_index()

        if hue is not None:
            if x is None and y is None:

                return data
            elif x is None:

                agg_dict = {}
                for hh in data[hue].unique():
                    agg_dict[hh] = data.loc[data[hue] == hh, y]
                df = mngs.pd.force_df(agg_dict)
                return df

            elif y is None:

                df = pd.concat(
                    [
                        data.loc[data[hue] == hh, x]
                        for hh in data[hue].unique()
                    ],
                    axis=1,
                )
                return df
            else:
                pivoted_data = data.pivot_table(
                    values=y,
                    index=data.index,
                    columns=[x, hue],
                    aggfunc="first",
                )
                pivoted_data.columns = [
                    f"{col[0]}-{col[1]}" for col in pivoted_data.columns
                ]
                return pivoted_data
        else:
            if x is None and y is None:
                return data

            elif x is None:
                return data[[y]]

            elif y is None:
                return data[[x]]

            else:
                return data.pivot_table(
                    values=y, index=data.index, columns=x, aggfunc="first"
                )

    @sns_copy_doc
    def sns_barplot(
        self, data=None, x=None, y=None, track=True, id=None, **kwargs
    ):
        self._sns_base_xyhue(
            "sns_barplot", data=data, x=x, y=y, track=track, id=id, **kwargs
        )

    @sns_copy_doc
    def sns_boxplot(
        self,
        data=None,
        x=None,
        y=None,
        strip=False,
        track=True,
        id=None,
        **kwargs,
    ):
        self._sns_base_xyhue(
            "sns_boxplot", data=data, x=x, y=y, track=track, id=id, **kwargs
        )
        if strip:
            strip_kwargs = kwargs.copy()
            strip_kwargs.pop("notch", None)  # Remove boxplot-specific kwargs
            strip_kwargs.pop("whis", None)
            self.sns_stripplot(
                data=data,
                x=x,
                y=y,
                track=False,
                id=f"{id}_strip",
                **strip_kwargs,
            )

    @sns_copy_doc
    def sns_heatmap(self, *args, xyz=False, track=True, id=None, **kwargs):
        method_name = "sns_heatmap"
        df = args[0]
        if xyz:
            df = mngs.pd.to_xyz(df)
        self._sns_base(
            method_name, *args, track=track, track_obj=df, id=id, **kwargs
        )

    @sns_copy_doc
    def sns_histplot(
        self, data=None, x=None, y=None, track=True, id=None, **kwargs
    ):
        self._sns_base_xyhue(
            "sns_histplot", data=data, x=x, y=y, track=track, id=id, **kwargs
        )

    @sns_copy_doc
    def sns_kdeplot(
        self,
        data=None,
        x=None,
        y=None,
        xlim=None,
        ylim=None,
        track=True,
        id=None,
        **kwargs,
    ):
        if kwargs.get("hue"):
            hue_col = kwargs["hue"]
            hues = data[hue_col]
            if x is not None:
                lim = xlim
                for hh in np.unique(hues):
                    _data = data.loc[data[hue_col] == hh, x]
                    self.kde(_data, xlim=lim, label=hh, id=hh, **kwargs)

            if y is not None:
                lim = xlim
                for hh in np.unique(hues):
                    _data = data.loc[data[hue] == hh, y]
                    self.kde(_data, xlim=lim, label=hh, id=hh, **kwargs)

        else:
            if x is not None:
                _data, lim = data[x], xlim
            if y is not None:
                _data, lim = data[y], ylim
            self.kde(_data, xlim=lim, **kwargs)

    @sns_copy_doc
    def sns_pairplot(self, *args, track=True, id=None, **kwargs):
        self._sns_base("sns_pairplot", *args, track=track, id=id, **kwargs)

    @sns_copy_doc
    def sns_scatterplot(
        self, data=None, x=None, y=None, track=True, id=None, **kwargs
    ):
        self._sns_base_xyhue(
            "sns_scatterplot",
            data=data,
            x=x,
            y=y,
            track=track,
            id=id,
            **kwargs,
        )

    @sns_copy_doc
    def sns_swarmplot(
        self, data=None, x=None, y=None, track=True, id=None, **kwargs
    ):
        self._sns_base_xyhue(
            "sns_swarmplot", data=data, x=x, y=y, track=track, id=id, **kwargs
        )

    @sns_copy_doc
    def sns_stripplot(
        self, data=None, x=None, y=None, track=True, id=None, **kwargs
    ):
        self._sns_base_xyhue(
            "sns_stripplot", data=data, x=x, y=y, track=track, id=id, **kwargs
        )

    # @sns_copy_doc
    # def sns_violinplot(
    #     self, data=None, x=None, y=None, track=True, id=None, **kwargs
    # ):
    #     self._sns_base_xyhue(
    #         "sns_violinplot", data=data, x=x, y=y, track=track, id=id, **kwargs
    #     )

    @sns_copy_doc
    def sns_violinplot(
        self,
        data=None,
        x=None,
        y=None,
        track=True,
        id=None,
        half=False,
        **kwargs,
    ):
        if half:
            with self._no_tracking():
                self.axis = mngs.plt.ax.half_violin(
                    self.axis, data=data, x=x, y=y, **kwargs
                )
        else:
            self._sns_base_xyhue(
                "sns_violinplot",
                data=data,
                x=x,
                y=y,
                track=track,
                id=id,
                **kwargs,
            )

        # Tracking
        track_obj = self._sns_prepare_xyhue(data, x, y, kwargs.get("hue"))
        self._track(track, id, "sns_violinplot", track_obj, kwargs)

        return self.axis

    # @sns_copy_doc
    # def sns_violinplot(
    #     self, data=None, x=None, y=None, track=True, id=None, half=False, **kwargs
    # ):
    #     if half:
    #         return self.sns_categorical_kde_plot(data=data, x=x, y=y, track=track, id=id, **kwargs)
    #     else:
    #         self._sns_base_xyhue(
    #             "sns_violinplot", data=data, x=x, y=y, track=track, id=id, **kwargs
    #         )

    # @sns_copy_doc
    # def sns_violinplot(
    #     self, data=None, x=None, y=None, track=True, id=None, half=False, **kwargs
    # ):
    #     if half:
    #         # Add a fake hue column
    #         data = data.copy()
    #         original_hue = kwargs.get('hue')
    #         data['_fake_hue'] = data[original_hue] + '_right'
    #         kwargs['hue'] = '_fake_hue'
    #         kwargs['split'] = True

    #         if 'hue_order' in kwargs:
    #             kwargs['hue_order'] = [h + '_right' for h in kwargs['hue_order']]

    #         if 'hue_colors' in kwargs:
    #             kwargs['hue_colors'] = kwargs['hue_colors'] * 2

    #     self._sns_base_xyhue(
    #         "sns_violinplot", data=data, x=x, y=y, track=track, id=id, **kwargs
    #     )

    # def sns_categorical_kde_plot(self, data=None, x=None, y=None, track=True, id=None, **kwargs):
    #     self._sns_base_xyhue(
    #         "sns_kdeplot", data=data, x=x, y=y, multiple="stack", track=track, id=id, **kwargs
    #     )

    @sns_copy_doc
    def sns_jointplot(self, *args, track=True, id=None, **kwargs):
        self._sns_base("sns_jointplot", *args, track=track, id=id, **kwargs)

    # @sns_copy_doc
    # def sns_half_violin(self, data=None, x=None, y=None, hue=None, track=True, id=None, **kwargs):
    #     # Method name
    #     method_name = "sns_half_violin"

    #     # Plotting
    #     with self._no_tracking():
    #         self.axis = mngs.plt.half_violin(self.axis, data=data, x=x, y=y, hue=hue, **kwargs)

    #     # Tracking
    #     track_obj = self._sns_prepare_xyhue(data, x, y, hue)
    #     self._track(track, id, method_name, track_obj, kwargs)

    #     return self.axis

    ################################################################################
    ## Adjusting methods
    ################################################################################
    def rotate_labels(self, x=30, y=30, x_ha="right", y_ha="center"):
        self.axis = mngs.plt.ax.rotate_labels(
            self.axis, x=x, y=y, x_ha=x_ha, y_ha=y_ha
        )

    def legend(self, loc="upper left"):
        return self.axis.legend(loc=loc)

    def set_xyt(
        self,
        x=None,
        y=None,
        t=None,
        format_labels=True,
    ):
        self.axis = mngs.plt.ax.set_xyt(
            self.axis,
            x=x,
            y=y,
            t=t,
            format_labels=format_labels,
        )

    def set_supxyt(
        self,
        xlabel=None,
        ylabel=None,
        title=None,
        format_labels=True,
    ):
        self.axis = mngs.plt.ax.set_supxyt(
            self.axis,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            format_labels=format_labels,
        )

    def set_ticks(
        self,
        xvals=None,
        xticks=None,
        yvals=None,
        yticks=None,
        **kwargs,
    ):

        self.axis = mngs.plt.ax.set_ticks(
            self.axis,
            xvals=xvals,
            xticks=xticks,
            yvals=yvals,
            yticks=yticks,
        )

    def set_n_ticks(self, n_xticks=4, n_yticks=4):
        self.axis = mngs.plt.ax.set_n_ticks(
            self.axis, n_xticks=n_xticks, n_yticks=n_yticks
        )

    def hide_spines(
        self,
        top=True,
        bottom=True,
        left=True,
        right=True,
        ticks=True,
        labels=True,
    ):
        self.axis = mngs.plt.ax.hide_spines(
            self.axis,
            top=top,
            bottom=bottom,
            left=left,
            right=right,
            ticks=ticks,
            labels=labels,
        )

    def extend(self, x_ratio=1.0, y_ratio=1.0):
        self.axis = mngs.plt.ax.extend(
            self.axis, x_ratio=x_ratio, y_ratio=y_ratio
        )

    def shift(self, dx=0, dy=0):
        self.axis = mngs.plt.ax.shift(self.axis, dx=dx, dy=dy)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-15 10:29:46 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_ChatGPT.py


"""
This script does XYZ.
"""


"""
Imports
"""
import sys

import matplotlib.pyplot as plt
import mngs

# from mngs.gen import notify
from openai import OpenAI

from ._BaseGenAI import BaseGenAI

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


class Perplexity(BaseGenAI):
    def __init__(
        self,
        system_setting="",
        model="",
        api_key="",
        stream=False,
        seed=None,
        n_keep=1,
        temperature=1.0,
        chat_history=None,
    ):
        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key=api_key,
            stream=stream,
            n_keep=n_keep,
            temperature=temperature,
            provider="Perplexity",
            chat_history=chat_history,
        )

    def _init_client(
        self,
    ):
        client = OpenAI(
            api_key=self.api_key, base_url="https://api.perplexity.ai"
        )
        return client

    def _api_call_static(self):
        output = self.client.chat.completions.create(
            model=self.model,
            messages=self.history,
            max_tokens=4096,
            stream=False,
            temperature=self.temperature,
            # return_citations=True,
        )

        out_text = output.choices[0].message.content
        self.input_tokens += output.usage.prompt_tokens
        self.output_tokens += output.usage.completion_tokens

        return out_text

    def _api_call_stream(self):
        stream = self.client.chat.completions.create(
            model=self.model,
            messages=self.history,
            max_tokens=4096,
            n=1,
            stream=self.stream,
            temperature=self.temperature,
            # return_citations=True,
        )

        for chunk in stream:
            if chunk:
                if chunk.choices[0].finish_reason == "stop":
                    try:
                        self.input_tokens += chunk.usage.prompt_tokens
                    except:
                        pass
                    try:
                        self.output_tokens += chunk.usage.completion_tokens
                    except:
                        pass

            if chunk.choices:
                current_text = chunk.choices[0].delta.content
                if current_text:
                    yield current_text

    def _get_available_models(self):
        return [
            "llama-3.1-sonar-small-128k-online",
            "llama-3.1-sonar-large-128k-online",
            "llama-3.1-sonar-huge-128k-online",
            "llama-3.1-sonar-small-128k-chat",
            "llama-3.1-sonar-large-128k-chat",
            "llama-3-sonar-small-32k-chat",
            "llama-3-sonar-small-32k-online",
            "llama-3-sonar-large-32k-chat",
            "llama-3-sonar-large-32k-online",
            "llama-3-8b-instruct",
            "llama-3-70b-instruct",
            "mixtral-8x7b-instruct",
        ]


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-06 10:47:42 (ywatanabe)"
# _fdr_correction.py

"""
Functionality:
    - Implements False Discovery Rate (FDR) correction for multiple comparisons
    - Provides both NumPy and PyTorch implementations

Input:
    - Array-like object of p-values
    - Alpha level for significance
    - Method for correction ('indep' or 'negcorr')

Output:
    - Boolean array indicating rejected hypotheses
    - Array of corrected p-values

Prerequisites:
    - NumPy, PyTorch, pandas, statsmodels, and mngs packages
"""

"""Imports"""
from typing import Union, Tuple
import numpy as np
import torch
import pandas as pd
from statsmodels.stats.multitest import fdrcorrection
import mngs
from mngs.gen import pandas_fn

ArrayLike = Union[np.ndarray, torch.Tensor, pd.Series]

@pandas_fn
def fdr_correction(results: pd.DataFrame) -> pd.DataFrame:
    """
    Apply FDR correction to p-value columns in a DataFrame.

    Example:
    --------
    >>> df = pd.DataFrame({'p_value': [0.01, 0.05, 0.1], 'other': [1, 2, 3]})
    >>> fdr_correction(df)
       p_value  other  p_value_fdr p_value_fdr_stars
    0     0.01      1    0.030000               **
    1     0.05      2    0.075000                *
    2     0.10      3    0.100000

    Parameters:
    -----------
    results : pd.DataFrame
        DataFrame containing p-value columns

    Returns:
    --------
    pd.DataFrame
        DataFrame with added FDR-corrected p-values and stars
    """
    pval_cols = mngs.stats.find_pval(results, multiple=True)
    if not pval_cols:
        return results

    for pval_col in pval_cols:
        non_nan = results.dropna(subset=[pval_col])
        nan_rows = results[results[pval_col].isna()]

        pvals = non_nan[pval_col]
        if isinstance(pvals, pd.DataFrame):
            pvals = pvals.values.flatten()

        _, fdr_corrected_pvals = fdrcorrection(pvals)
        non_nan[f"{pval_col}_fdr"] = fdr_corrected_pvals
        nan_rows[f"{pval_col}_fdr"] = np.nan

        results = pd.concat([non_nan, nan_rows]).sort_index()
        results[f"{pval_col}_fdr_stars"] = results[f"{pval_col}_fdr"].apply(mngs.stats.p2stars)

    return results

# def _ecdf(xx: ArrayLike) -> ArrayLike:
#     """Compute empirical cumulative distribution function."""
#     nobs = len(xx)
#     return np.arange(1, nobs + 1) / float(nobs)

# def _ecdf_torch(xx: torch.Tensor) -> torch.Tensor:
#     """Compute empirical cumulative distribution function using PyTorch."""
#     nobs = len(xx)
#     return torch.arange(1, nobs + 1, device=xx.device) / float(nobs)

# def fdr_correction_torch(pvals: torch.Tensor, alpha: float = 0.05, method: str = "indep") -> Tuple[torch.Tensor, torch.Tensor]:
#     """
#     P-value correction with False Discovery Rate (FDR) using PyTorch.

#     Example:
#     >>> pvals = torch.tensor([0.01, 0.02, 0.03, 0.04, 0.05])
#     >>> reject, pvals_corrected = fdr_correction_torch(pvals)
#     >>> print(reject, pvals_corrected)

#     Parameters:
#     -----------
#     pvals : torch.Tensor
#         Set of p-values of the individual tests
#     alpha : float, optional
#         Error rate (default is 0.05)
#     method : str, optional
#         'indep' for Benjamini/Hochberg, 'negcorr' for Benjamini/Yekutieli (default is 'indep')

#     Returns:
#     --------
#     reject : torch.Tensor
#         Boolean tensor indicating rejected hypotheses
#     pvals_corrected : torch.Tensor
#         Tensor of corrected p-values
#     """
#     shape_init = pvals.shape
#     pvals = pvals.ravel()

#     pvals_sortind = torch.argsort(pvals)
#     pvals_sorted = pvals[pvals_sortind]
#     sortrevind = pvals_sortind.argsort()

#     if method in ["i", "indep", "p", "poscorr"]:
#         ecdffactor = _ecdf_torch(pvals_sorted)
#     elif method in ["n", "negcorr"]:
#         cm = torch.sum(1.0 / torch.arange(1, len(pvals_sorted) + 1, device=pvals.device))
#         ecdffactor = _ecdf_torch(pvals_sorted) / cm
#     else:
#         raise ValueError("Method should be 'indep' or 'negcorr'")

#     ecdffactor = ecdffactor.to(pvals_sorted.dtype)

#     reject = pvals_sorted < (ecdffactor * alpha)

#     if reject.any():
#         rejectmax = torch.nonzero(reject, as_tuple=True)[0].max()
#     else:
#         rejectmax = torch.tensor(0, device=pvals.device)
#     reject[:rejectmax+1] = True

#     pvals_corrected_raw = pvals_sorted / ecdffactor
#     pvals_corrected = torch.minimum(torch.ones_like(pvals_corrected_raw), torch.cummin(pvals_corrected_raw.flip(0), 0)[0].flip(0))

#     pvals_corrected = pvals_corrected[sortrevind].reshape(shape_init)
#     reject = reject[sortrevind].reshape(shape_init)
#     return reject, pvals_corrected

if __name__ == "__main__":
    pvals = [0.02, 0.03, 0.05]
    pvals_torch = torch.tensor(np.array([0.02, 0.03, 0.05]))

    reject, pvals_corrected = fdr_correction(pd.DataFrame({'p_value': pvals}))

    reject_torch, pvals_corrected_torch = fdr_correction_torch(
        pvals_torch, alpha=0.05, method="indep"
    )

    arr = pvals_corrected['fdr_p_value'].to_numpy().astype(float)
    tor = pvals_corrected_torch.numpy().astype(float)
    print(mngs.general.isclose(arr, tor))

# #!/usr/bin/env python3
# # Time-stamp: "2024-10-06 09:26:33 (ywatanabe)"

# """
# Functionality:
#     - Implements False Discovery Rate (FDR) correction for multiple comparisons
#     - Provides both NumPy and PyTorch implementations

# Input:
#     - Array-like object of p-values
#     - Alpha level for significance
#     - Method for correction ('indep' or 'negcorr')

# Output:
#     - Boolean array indicating rejected hypotheses
#     - Array of corrected p-values

# Prerequisites:
#     - NumPy, PyTorch, pandas, statsmodels, and mngs packages
# """

# """Imports"""
# from typing import Union, Tuple
# import numpy as np
# import torch
# import pandas as pd
# from statsmodels.stats.multitest import fdrcorrection
# import mngs

# ArrayLike = Union[np.ndarray, torch.Tensor, pd.Series]

# def fdr_correction(results: pd.DataFrame) -> pd.DataFrame:
#     """
#     Apply FDR correction to p-values in a DataFrame.

#     Example:
#     >>> df = pd.DataFrame({'p_value': [0.01, 0.02, 0.03, 0.04, 0.05]})
#     >>> corrected_df = fdr_correction(df)
#     >>> print(corrected_df)

#     Parameters:
#     -----------
#     results : pd.DataFrame
#         DataFrame containing a 'p_value' column

#     Returns:
#     --------
#     pd.DataFrame
#         DataFrame with added 'fdr_p_value' and 'fdr_stars' columns
#     """
#     if "p_value" not in results.columns:
#         return results
#     _, fdr_corrected_pvals = fdrcorrection(results["p_value"])
#     results["fdr_p_value"] = fdr_corrected_pvals
#     results["fdr_stars"] = results["fdr_p_value"].apply(mngs.stats.p2stars)
#     return results

# def fdr_correction(pvals, alpha=0.05, method="indep"):
#     # https://github.com/mne-tools/mne-python/blob/main/mne/stats/multi_comp.py
#     """P-value correction with False Discovery Rate (FDR).

#     Correction for multiple comparison using FDR :footcite:`GenoveseEtAl2002`.

#     This covers Benjamini/Hochberg for independent or positively correlated and
#     Benjamini/Yekutieli for general or negatively correlated tests.

#     Parameters
#     ----------
#     pvals : array_like
#         Set of p-values of the individual tests.
#     alpha : float
#         Error rate.
#     method : 'indep' | 'negcorr'
#         If 'indep' it implements Benjamini/Hochberg for independent or if
#         'negcorr' it corresponds to Benjamini/Yekutieli.

#     Returns
#     -------
#     reject : array, bool
#         True if a hypothesis is rejected, False if not.
#     pval_corrected : array
#     alpha : float
#         Error rate.
#     method : 'indep' | 'negcorr'
#         If 'indep' it implements Benjamini/Hochberg for independent or if
#         'negcorr' it corresponds to Benjamini/Yekutieli.

#     Returns
#     -------
#     reject : array, bool
#         True if a hypothesis is rejected, False if not.
#     pval_corrected : array
#     alpha : float
#         Error rate.
#     method : 'indep' | 'negcorr'
#         If 'indep' it implements Benjamini/Hochberg for independent or if
#         'negcorr' it corresponds to Benjamini/Yekutieli.

#     Returns
#     -------
#     reject : array, bool
#         True if a hypothesis is rejected, False if not.
#     pval_corrected : array
#         P-values adjusted for multiple hypothesis testing to limit FDR.

#     Returns
#     -------
#     reject : array, bool
#         True if a hypothesis is rejected, False if not.
#     pval_corrected : array
#         P-values adjusted for multiple hypothesis testing to limit FDR.

#     References
#     ----------
#     .. footbibliography::
#     """
#     pvals = np.asarray(pvals)
#     shape_init = pvals.shape
#     pvals = pvals.ravel()

#     pvals_sortind = np.argsort(pvals)
#     pvals_sorted = pvals[pvals_sortind]
#     sortrevind = pvals_sortind.argsort()

#     if method in ["i", "indep", "p", "poscorr"]:
#         ecdffactor = _ecdf(pvals_sorted)
#     elif method in ["n", "negcorr"]:
#         cm = np.sum(1.0 / np.arange(1, len(pvals_sorted) + 1))
#         ecdffactor = _ecdf(pvals_sorted) / cm
#     else:
#         raise ValueError("Method should be 'indep' and 'negcorr'")

#     reject = pvals_sorted < (ecdffactor * alpha)
#     if reject.any():
#         rejectmax = max(np.nonzero(reject)[0])
#     else:
#         rejectmax = 0
#     reject[:rejectmax] = True

#     pvals_corrected_raw = pvals_sorted / ecdffactor
#     pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
#     pvals_corrected[pvals_corrected > 1.0] = 1.0
#     pvals_corrected = pvals_corrected[sortrevind].reshape(shape_init)
#     reject = reject[sortrevind].reshape(shape_init)
#     return reject, pvals_corrected


# def fdr_correction(results: pd.DataFrame) -> pd.DataFrame:
#     if "p_value" not in results.columns:
#         return results
#     _, fdr_corrected_pvals = fdrcorrection(results["p_value"])
#     results["fdr_p_value"] = fdr_corrected_pvals
#     results["fdr_stars"] = results["fdr_p_value"].apply(mngs.stats.p2stars)
#     return results

# def _ecdf(xx: ArrayLike) -> ArrayLike:
#     """Compute empirical cumulative distribution function."""
#     nobs = len(xx)
#     return np.arange(1, nobs + 1) / float(nobs)

# def _ecdf_torch(xx: torch.Tensor) -> torch.Tensor:
#     """Compute empirical cumulative distribution function using PyTorch."""
#     nobs = len(xx)
#     return torch.arange(1, nobs + 1, device=xx.device) / float(nobs)

# def fdr_correction_torch(pvals: torch.Tensor, alpha: float = 0.05, method: str = "indep") -> Tuple[torch.Tensor, torch.Tensor]:
#     """
#     P-value correction with False Discovery Rate (FDR) using PyTorch.

#     Example:
#     >>> pvals = torch.tensor([0.01, 0.02, 0.03, 0.04, 0.05])
#     >>> reject, pvals_corrected = fdr_correction_torch(pvals)
#     >>> print(reject, pvals_corrected)

#     Parameters:
#     -----------
#     pvals : torch.Tensor
#         Set of p-values of the individual tests
#     alpha : float, optional
#         Error rate (default is 0.05)
#     method : str, optional
#         'indep' for Benjamini/Hochberg, 'negcorr' for Benjamini/Yekutieli (default is 'indep')

#     Returns:
#     --------
#     reject : torch.Tensor
#         Boolean tensor indicating rejected hypotheses
#     pvals_corrected : torch.Tensor
#         Tensor of corrected p-values
#     """
#     shape_init = pvals.shape
#     pvals = pvals.ravel()

#     pvals_sortind = torch.argsort(pvals)
#     pvals_sorted = pvals[pvals_sortind]
#     sortrevind = pvals_sortind.argsort()

#     if method in ["i", "indep", "p", "poscorr"]:
#         ecdffactor = _ecdf_torch(pvals_sorted)
#     elif method in ["n", "negcorr"]:
#         cm = torch.sum(1.0 / torch.arange(1, len(pvals_sorted) + 1, device=pvals.device))
#         ecdffactor = _ecdf_torch(pvals_sorted) / cm
#     else:
#         raise ValueError("Method should be 'indep' or 'negcorr'")

#     ecdffactor = ecdffactor.to(pvals_sorted.dtype)

#     reject = pvals_sorted < (ecdffactor * alpha)

#     if reject.any():
#         rejectmax = torch.nonzero(reject, as_tuple=True)[0].max()
#     else:
#         rejectmax = torch.tensor(0, device=pvals.device)
#     reject[:rejectmax+1] = True

#     pvals_corrected_raw = pvals_sorted / ecdffactor
#     pvals_corrected = torch.minimum(torch.ones_like(pvals_corrected_raw), torch.cummin(pvals_corrected_raw.flip(0), 0)[0].flip(0))

#     pvals_corrected = pvals_corrected[sortrevind].reshape(shape_init)
#     reject = reject[sortrevind].reshape(shape_init)
#     return reject, pvals_corrected


# if __name__ == "__main__":
#     pvals = [0.02, 0.03, 0.05]
#     pvals_torch = torch.tensor(np.array([0.02, 0.03, 0.05]))

#     reject, pvals_corrected = fdr_correction(pvals, alpha=0.05, method="indep")

#     reject_torch, pvals_corrected_torch = fdr_correction_torch(
#         pvals, alpha=0.05, method="indep"
#     )

#     arr = pvals_corrected.astype(float)
#     tor = pvals_corrected_torch.numpy().astype(float)
#     print(mngs.general.isclose(arr, tor))



#!/usr/bin/env python3

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def annotated_heatmap(
    cm,
    labels=None,
    title=None,
    cmap="Blues",
    norm=None,
    xlabel=None,
    ylabel=None,
    annot=True,
):
    df = pd.DataFrame(data=cm)

    if labels is not None:
        df.index = labels
        df.columns = labels

    fig, ax = plt.subplots()
    res = sns.heatmap(
        df,
        annot=annot,
        fmt=".3f",
        cmap=cmap,
        norm=norm,
    )  # cbar_kws={"shrink": 0.82}
    res.invert_yaxis()

    # make the frame invisible
    for _, spine in res.spines.items():
        spine.set_visible(False)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    return fig


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib import colors

    labels = ["T2T", "F2T", "T2F", "F2F"]
    arr = np.random.randint(0, 10, len(labels) ** 2).reshape(len(labels), len(labels))

    ## quantized, arbitrary range colormap you want
    cmap = colors.ListedColormap(
        ["navy", "royalblue", "lightsteelblue", "beige"],
    )
    norm = colors.BoundaryNorm([2, 4, 6, 8], cmap.N - 1)

    fig = annotated_heatmap(arr, cmap=cmap, norm=norm, labels=labels)
    fig.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-26 07:22:18 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/pd/_from_xyz.py

import pandas as pd
import numpy as np

def from_xyz(data_frame, x=None, y=None, z=None):
    """
    Convert a DataFrame with 'x', 'y', 'z' format into a heatmap DataFrame.

    Example
    -------
    import pandas as pd
    data = pd.DataFrame({
        'col1': ['A', 'B', 'C', 'A'],
        'col2': ['X', 'Y', 'Z', 'Y'],
        'p_val': [0.01, 0.05, 0.001, 0.1]
    })
    data = data.rename(columns={"col1": "x", "col2": "y", "p_val": "z"})
    result = from_xyz(data)
    print(result)

    Parameters
    ----------
    data_frame : pandas.DataFrame
        Input DataFrame with columns for x, y, and z values.
    x : str, optional
        Name of the column to use as x-axis. Defaults to 'x'.
    y : str, optional
        Name of the column to use as y-axis. Defaults to 'y'.
    z : str, optional
        Name of the column to use as z-values. Defaults to 'z'.

    Returns
    -------
    pandas.DataFrame
        A square DataFrame in heatmap format.
    """
    x = x or 'x'
    y = y or 'y'
    z = z or 'z'

    heatmap = pd.pivot_table(data_frame, values=z, index=y, columns=x, aggfunc='first')

    all_labels = sorted(set(heatmap.index) | set(heatmap.columns))
    heatmap = heatmap.reindex(index=all_labels, columns=all_labels)

    heatmap = heatmap.fillna(0)

    return heatmap

if __name__ == '__main__':
    np.random.seed(42)
    stats = pd.DataFrame({
        'col1': np.random.choice(['A', 'B', 'C'], 100),
        'col2': np.random.choice(['X', 'Y', 'Z'], 100),
        'p_val': np.random.rand(100)
    })
    stats = stats.rename(columns={"col1": "x", "col2": "y", "p_val": "z"})
    result = from_xyz(stats)
    print(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2021-09-20 21:02:04 (ywatanabe)"

import resource
import mngs


def limit_RAM(RAM_factor):
    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    max_val = min(RAM_factor * get_RAM() * 1024, get_RAM() * 1024)
    resource.setrlimit(resource.RLIMIT_AS, (max_val, hard))
    print(f"\nFree RAM was limited to {mngs.general.fmt_size(max_val)}")


def get_RAM():
    with open("/proc/meminfo", "r") as mem:
        free_memory = 0
        for i in mem:
            sline = i.split()
            if str(sline[0]) in ("MemFree:", "Buffers:", "Cached:"):
                free_memory += int(sline[1])
    return free_memory


if __name__ == "__main__":
    get_RAM()
    limit_RAM(0.1)
